@@ -9,6 +9,7 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 #include "pyodbc.h"
+#include "buffer.h"
 #include "wrapper.h"
 #include "textenc.h"
 #include "connection.h"
@@ -16,15 +17,6 @@
 #include "pyodbcmodule.h"
 #include "errors.h"
 #include "cnxninfo.h"
-#include "sqlwchar.h"
-
-#ifdef WORDS_BIGENDIAN
-# define OPTENC_UTF16NE OPTENC_UTF16BE
-# define ENCSTR_UTF16NE "utf-16be"
-#else
-# define OPTENC_UTF16NE OPTENC_UTF16LE
-# define ENCSTR_UTF16NE "utf-16le"
-#endif
 
 #if PY_MAJOR_VERSION < 3
 static bool IsStringType(PyObject* t) { return (void*)t == (void*)&PyString_Type; }
@@ -63,14 +55,6 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeou
     // This should have been checked by the global connect function.
     I(PyString_Check(pConnectString) || PyUnicode_Check(pConnectString));
 
-    const int cchMax = 600;
-
-    if (PySequence_Length(pConnectString) >= cchMax)
-    {
-        PyErr_SetString(PyExc_TypeError, "connection string too long");
-        return false;
-    }
-
     // The driver manager determines if the app is a Unicode app based on whether we call SQLDriverConnectA or
     // SQLDriverConnectW.  Some drivers, notably Microsoft Access/Jet, change their behavior based on this, so we try
     // the Unicode version first.  (The Access driver only supports Unicode text, but SQLDescribeCol returns SQL_CHAR
@@ -92,26 +76,53 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeou
             RaiseErrorFromHandle(0, "SQLSetConnectAttr(SQL_ATTR_LOGIN_TIMEOUT)", hdbc, SQL_NULL_HANDLE);
     }
 
+    const char* szEncoding = 0;
+    Object encBytes;
+    if (encoding)
+    {
+        #if PY_MAJOR_VERSION < 3
+        if (PyString_Check(encoding))
+        {
+            szEncoding = PyString_AsString(encoding);
+            if (!szEncoding)
+                return false;
+        }
+        #endif
+        if (PyUnicode_Check(encoding))
+        {
+            #if PY_MAJOR_VERSION < 3
+                encBytes = PyUnicode_AsUTF8String(encoding);
+                if (!encBytes)
+                    return false;
+                szEncoding = PyString_AS_STRING(encBytes.Get());
+            #else
+                szEncoding = PyUnicode_AsUTF8(encoding);
+            #endif
+        }
+    }
+
     if (!fAnsi)
     {
         // I want to call the W version when possible since the driver can use it as an
-        // indication that we can handle Unicode.  We are going to use the same unicode ending
-        // as we do for binding parameters.
+        // indication that we can handle Unicode.
 
-        SQLWChar wchar(pConnectString, SQL_C_WCHAR, encoding, ENCSTR_UTF16NE);
-        if (!wchar)
+        SQLWChar wchar(pConnectString, szEncoding ? szEncoding : ENCSTR_UTF16NE);
+        if (!wchar.isValid())
             return false;
 
         Py_BEGIN_ALLOW_THREADS
-        ret = SQLDriverConnectW(hdbc, 0, (SQLWCHAR*)wchar.value(), (SQLSMALLINT)wchar.charlen(), 0, 0, 0, SQL_DRIVER_NOPROMPT);
+        ret = SQLDriverConnectW(hdbc, 0, wchar.psz, SQL_NTS, 0, 0, 0, SQL_DRIVER_NOPROMPT);
         Py_END_ALLOW_THREADS
         if (SQL_SUCCEEDED(ret))
             return true;
     }
 
-    SQLWChar ch(pConnectString, SQL_C_CHAR, encoding, "utf-8");
+    SQLWChar wchar(pConnectString, szEncoding ? szEncoding : "utf-8");
+    if (!wchar.isValid())
+        return false;
+
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLDriverConnect(hdbc, 0, (SQLCHAR*)ch.value(), (SQLSMALLINT)ch.charlen(), 0, 0, 0, SQL_DRIVER_NOPROMPT);
+    ret = SQLDriverConnect(hdbc, 0, (SQLCHAR*)wchar.psz, SQL_NTS, 0, 0, 0, SQL_DRIVER_NOPROMPT);
     Py_END_ALLOW_THREADS
     if (SQL_SUCCEEDED(ret))
         return true;
@@ -121,6 +132,102 @@ static bool Connect(PyObject* pConnectString, HDBC hdbc, bool fAnsi, long timeou
     return false;
 }
 
+static bool ApplyPreconnAttrs(HDBC hdbc, SQLINTEGER ikey, PyObject *value, char *strencoding)
+{
+    SQLRETURN ret;
+    SQLPOINTER ivalue = 0;
+    SQLINTEGER vallen = 0;
+
+    if (PyLong_Check(value))
+    {
+        if (_PyLong_Sign(value) >= 0)
+        {
+            ivalue = (SQLPOINTER)PyLong_AsUnsignedLong(value);
+            vallen = SQL_IS_UINTEGER;
+        } else
+        {
+            ivalue = (SQLPOINTER)PyLong_AsLong(value);
+            vallen = SQL_IS_INTEGER;
+        }
+    }
+#if PY_MAJOR_VERSION < 3
+    else if (PyInt_Check(value))
+    {
+        ivalue = (SQLPOINTER)PyInt_AsLong(value);
+        vallen = SQL_IS_INTEGER;
+    }
+    else if (PyBuffer_Check(value))
+    {
+        // We can only assume and take the first segment.
+        PyBuffer_GetMemory(value, (const char**)&ivalue);
+        vallen = SQL_IS_POINTER;
+    }
+#endif
+#if PY_VERSION_HEX >= 0x02060000
+    else if (PyByteArray_Check(value))
+    {
+        ivalue = (SQLPOINTER)PyByteArray_AsString(value);
+        vallen = SQL_IS_POINTER;
+    }
+#endif
+    else if (PyBytes_Check(value))
+    {
+        ivalue = PyBytes_AS_STRING(value);
+#if PY_MAJOR_VERSION < 3
+        vallen = SQL_NTS;
+#else
+        vallen = SQL_IS_POINTER;
+#endif
+    }
+    else if (PyUnicode_Check(value))
+    {
+        Object stringholder;
+if (sizeof(Py_UNICODE) == 2 // This part should be compile-time.
+    && (!strencoding || !strcmp(strencoding, "utf-16le")))
+{
+        // default or utf-16le is set, pass through directly
+        ivalue = PyUnicode_AS_UNICODE(value);
+}
+else
+{
+        // use strencoding to convert, default to utf-16le if not set.
+        stringholder = PyCodec_Encode(value, strencoding ? strencoding : "utf-16le", "strict");
+        ivalue = PyBytes_AS_STRING(stringholder.Get());
+}
+        vallen = SQL_NTS;
+        Py_BEGIN_ALLOW_THREADS
+        ret = SQLSetConnectAttrW(hdbc, ikey, ivalue, vallen);
+        Py_END_ALLOW_THREADS
+        goto checkSuccess;
+    }
+    else if (PySequence_Check(value))
+    {
+        // To allow for possibility of setting multiple attributes more than once.
+        Py_ssize_t len = PySequence_Size(value);
+        for (Py_ssize_t i = 0; i < len; i++)
+        {
+            Object v(PySequence_GetItem(value, i));
+            if (!ApplyPreconnAttrs(hdbc, ikey, v.Get(), strencoding))
+                return false;
+        }
+        return true;
+    }
+
+    Py_BEGIN_ALLOW_THREADS
+    ret = SQLSetConnectAttr(hdbc, ikey, ivalue, vallen);
+    Py_END_ALLOW_THREADS
+
+checkSuccess:
+    if (!SQL_SUCCEEDED(ret))
+    {
+        RaiseErrorFromHandle(0, "SQLSetConnectAttr", hdbc, SQL_NULL_HANDLE);
+        Py_BEGIN_ALLOW_THREADS
+        SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
+        Py_END_ALLOW_THREADS
+        return false;
+    }
+    return true;
+}
 
 PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi, long timeout, bool fReadOnly,
                          PyObject* attrs_before, Object& encoding)
@@ -135,6 +242,7 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     // Allocate HDBC and connect
     //
 
+    Object attrs_before_o(attrs_before);
     HDBC hdbc = SQL_NULL_HANDLE;
     SQLRETURN ret;
     Py_BEGIN_ALLOW_THREADS
@@ -152,29 +260,24 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
         Py_ssize_t pos = 0;
         PyObject* key = 0;
         PyObject* value = 0;
+
+        Object encodingholder;
+        char *strencoding = encoding.Get() ?
+            (PyUnicode_Check(encoding) ? PyBytes_AsString(encodingholder = PyCodec_Encode(encoding, "utf-8", "strict")) :
+             PyBytes_Check(encoding) ? PyBytes_AsString(encoding) : 0) : 0;
+
         while (PyDict_Next(attrs_before, &pos, &key, &value))
         {
-            int ikey = 0, ivalue = 0;
-#if PY_MAJOR_VERSION < 3
-            if (PyInt_Check(key))
-                ikey = (int)PyInt_AsLong(key);
-            if (PyInt_Check(value))
-                ivalue = (int)PyInt_AsLong(value);
-#endif
+            SQLINTEGER ikey = 0;
+
             if (PyLong_Check(key))
                 ikey = (int)PyLong_AsLong(key);
-            if (PyLong_Check(value))
-                ivalue = (int)PyLong_AsLong(value);
-
-            Py_BEGIN_ALLOW_THREADS
-            ret = SQLSetConnectAttr(hdbc, ikey, (SQLPOINTER)(uintptr_t)ivalue, SQL_IS_INTEGER);
-            Py_END_ALLOW_THREADS
-            if (!SQL_SUCCEEDED(ret))
+#if PY_MAJOR_VERSION < 3
+            else if (PyInt_Check(key))
+                ikey = (int)PyInt_AsLong(key);
+#endif
+            if (!ApplyPreconnAttrs(hdbc, ikey, value, strencoding))
             {
-                RaiseErrorFromHandle(0, "SQLSetConnectAttr", hdbc, SQL_NULL_HANDLE);
-                Py_BEGIN_ALLOW_THREADS
-                SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
-                Py_END_ALLOW_THREADS
                 return 0;
             }
         }
@@ -220,6 +323,8 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
     cnxn->conv_types   = 0;
     cnxn->conv_funcs   = 0;
 
+    cnxn->attrs_before = attrs_before_o.Detach();
+
     // This is an inefficient default, but should work all the time.  When we are offered
     // single-byte text we don't actually know what the encoding is.  For example, with SQL
     // Server the encoding is based on the database's collation.  We ask the driver / DB to
@@ -281,7 +386,7 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
 
         if (!SQL_SUCCEEDED(ret))
         {
-            RaiseErrorFromHandle(cnxn, "SQLSetConnnectAttr(SQL_ATTR_AUTOCOMMIT)", cnxn->hdbc, SQL_NULL_HANDLE);
+            RaiseErrorFromHandle(cnxn, "SQLSetConnectAttr(SQL_ATTR_AUTOCOMMIT)", cnxn->hdbc, SQL_NULL_HANDLE);
             Py_DECREF(cnxn);
             return 0;
         }
@@ -295,7 +400,7 @@ PyObject* Connection_New(PyObject* pConnectString, bool fAutoCommit, bool fAnsi,
 
         if (!SQL_SUCCEEDED(ret))
         {
-            RaiseErrorFromHandle(cnxn, "SQLSetConnnectAttr(SQL_ATTR_ACCESS_MODE)", cnxn->hdbc, SQL_NULL_HANDLE);
+            RaiseErrorFromHandle(cnxn, "SQLSetConnectAttr(SQL_ATTR_ACCESS_MODE)", cnxn->hdbc, SQL_NULL_HANDLE);
             Py_DECREF(cnxn);
             return 0;
         }
@@ -423,6 +528,9 @@ static int Connection_clear(PyObject* self)
     cnxn->str_enc.name = 0;
 #endif
 
+    Py_XDECREF(cnxn->attrs_before);
+    cnxn->attrs_before = 0;
+
     _clear_conv(cnxn);
 
     return 0;
@@ -504,6 +612,35 @@ struct GetInfoType
 };
 
 static const GetInfoType aInfoTypes[] = {
+
+    // SQL_CONVERT_X
+    { SQL_CONVERT_FUNCTIONS, GI_UINTEGER },
+    { SQL_CONVERT_BIGINT, GI_UINTEGER },
+    { SQL_CONVERT_BINARY, GI_UINTEGER },
+    { SQL_CONVERT_BIT, GI_UINTEGER },
+    { SQL_CONVERT_CHAR, GI_UINTEGER },
+    { SQL_CONVERT_DATE, GI_UINTEGER },
+    { SQL_CONVERT_DECIMAL, GI_UINTEGER },
+    { SQL_CONVERT_DOUBLE, GI_UINTEGER },
+    { SQL_CONVERT_FLOAT, GI_UINTEGER },
+    { SQL_CONVERT_INTEGER, GI_UINTEGER },
+    { SQL_CONVERT_LONGVARCHAR, GI_UINTEGER },
+    { SQL_CONVERT_NUMERIC, GI_UINTEGER },
+    { SQL_CONVERT_REAL, GI_UINTEGER },
+    { SQL_CONVERT_SMALLINT, GI_UINTEGER },
+    { SQL_CONVERT_TIME, GI_UINTEGER },
+    { SQL_CONVERT_TIMESTAMP, GI_UINTEGER },
+    { SQL_CONVERT_TINYINT, GI_UINTEGER },
+    { SQL_CONVERT_VARBINARY, GI_UINTEGER },
+    { SQL_CONVERT_VARCHAR, GI_UINTEGER },
+    { SQL_CONVERT_LONGVARBINARY, GI_UINTEGER },
+    { SQL_CONVERT_WCHAR, GI_UINTEGER },
+    { SQL_CONVERT_INTERVAL_DAY_TIME, GI_UINTEGER },
+    { SQL_CONVERT_INTERVAL_YEAR_MONTH, GI_UINTEGER },
+    { SQL_CONVERT_WLONGVARCHAR, GI_UINTEGER },
+    { SQL_CONVERT_WVARCHAR, GI_UINTEGER },
+    { SQL_CONVERT_GUID, GI_UINTEGER },
+
     { SQL_ACCESSIBLE_PROCEDURES, GI_YESNO },
     { SQL_ACCESSIBLE_TABLES, GI_YESNO },
     { SQL_ACTIVE_ENVIRONMENTS, GI_USMALLINT },
@@ -522,8 +659,6 @@ static const GetInfoType aInfoTypes[] = {
     { SQL_COLLATION_SEQ, GI_STRING },
     { SQL_COLUMN_ALIAS, GI_YESNO },
     { SQL_CONCAT_NULL_BEHAVIOR, GI_USMALLINT },
-    { SQL_CONVERT_FUNCTIONS, GI_UINTEGER },
-    { SQL_CONVERT_VARCHAR, GI_UINTEGER },
     { SQL_CORRELATION_NAME, GI_USMALLINT },
     { SQL_CREATE_ASSERTION, GI_UINTEGER },
     { SQL_CREATE_CHARACTER_SET, GI_UINTEGER },
@@ -948,6 +1083,54 @@ static int Connection_settimeout(PyObject* self, PyObject* value, void* closure)
     return 0;
 }
 
+static bool _remove_converter(PyObject* self, SQLSMALLINT sqltype)
+{
+    Connection* cnxn = (Connection*)self;
+
+    if (!cnxn->conv_count)
+    {
+        // There are no converters, so nothing to remove.
+        return true;
+    }
+
+    int          count = cnxn->conv_count;
+    SQLSMALLINT* types = cnxn->conv_types;
+    PyObject**   funcs = cnxn->conv_funcs;
+
+    int i = 0;
+    for (; i < count; i++)
+        if (types[i] == sqltype)
+            break;
+
+    if (i == count)
+    {
+        // There is no converter for this type, so nothing to remove.
+        return true;
+    }
+
+    Py_DECREF(funcs[i]);
+
+    int move = count - i - 1;  // How many are we moving?
+    if (move > 0)
+    {
+        memcpy(&types[i], &types[i+1], move * sizeof(SQLSMALLINT));
+        memcpy(&funcs[i], &funcs[i+1], move * sizeof(PyObject*));
+    }
+    count--;
+
+    // Note: If the realloc fails, the old array is still around and is 1 element too long but
+    // everything will still work, so we ignore.
+    pyodbc_realloc((BYTE**)&types, count * sizeof(SQLSMALLINT));
+    pyodbc_realloc((BYTE**)&funcs, count * sizeof(PyObject*));
+
+    cnxn->conv_count = count;
+    cnxn->conv_types = types;
+    cnxn->conv_funcs = funcs;
+
+    return true;
+}
+
+
 static bool _add_converter(PyObject* self, SQLSMALLINT sqltype, PyObject* func)
 {
     Connection* cnxn = (Connection*)self;
@@ -1022,10 +1205,12 @@ static char conv_add_doc[] =
     "  value, and should return the converted value.  If the value is NULL, the\n"
     "  parameter will be None.  Otherwise it will be a "
 #if PY_MAJOR_VERSION >= 3
-    "bytes object."
+    "bytes object.\n"
 #else
-    "str object with the raw bytes."
+    "str object with the raw bytes.\n"
 #endif
+    "\n"
+    "If func is None, any existing converter is removed."
     ;
 
 static PyObject* Connection_conv_add(PyObject* self, PyObject* args)
@@ -1035,12 +1220,84 @@ static PyObject* Connection_conv_add(PyObject* self, PyObject* args)
     if (!PyArg_ParseTuple(args, "iO", &sqltype, &func))
         return 0;
 
-    if (!_add_converter(self, (SQLSMALLINT)sqltype, func))
+    if (func != Py_None)
+    {
+        if (!_add_converter(self, (SQLSMALLINT)sqltype, func))
+            return 0;
+    }
+    else
+    {
+        if (!_remove_converter(self, (SQLSMALLINT)sqltype))
+            return 0;
+    }
+
+    Py_RETURN_NONE;
+}
+
+static char conv_remove_doc[] =
+    "remove_output_converter(sqltype) --> None\n"
+    "\n"
+    "Remove an output converter function that was registered with\n"
+    "add_output_converter.  It is safe to call if no converter is\n"
+    "registered for the type.\n"
+    "\n"
+    "sqltype\n"
+    "  The integer SQL type value being converted, which can be one of the defined\n"
+    "  standard constants (e.g. pyodbc.SQL_VARCHAR) or a database-specific value\n"
+    "  (e.g. -151 for the SQL Server 2008 geometry data type).\n"
+    ;
+
+static PyObject* Connection_conv_remove(PyObject* self, PyObject* args)
+{
+    int sqltype;
+    if (!PyArg_ParseTuple(args, "i", &sqltype))
+        return 0;
+
+    if (!_remove_converter(self, (SQLSMALLINT)sqltype))
         return 0;
 
     Py_RETURN_NONE;
 }
 
+static char conv_get_doc[] =
+    "get_output_converter(sqltype) --> <class 'function'>\n"
+    "\n"
+    "Get the output converter function that was registered with\n"
+    "add_output_converter.  It is safe to call if no converter is\n"
+    "registered for the type (returns None).\n"
+    "\n"
+    "sqltype\n"
+    "  The integer SQL type value being converted, which can be one of the defined\n"
+    "  standard constants (e.g. pyodbc.SQL_VARCHAR) or a database-specific value\n"
+    "  (e.g. -151 for the SQL Server 2008 geometry data type).\n"
+    ;
+
+static PyObject* _get_converter(PyObject* self, SQLSMALLINT sqltype)
+{
+    Connection* cnxn = (Connection*)self;
+
+    if (cnxn->conv_count)
+    {
+        for (int i = 0; i < cnxn->conv_count; i++)
+        {
+            if (cnxn->conv_types[i] == sqltype)
+            {
+                return cnxn->conv_funcs[i];
+            }
+        }
+    }
+    Py_RETURN_NONE;
+}
+
+static PyObject* Connection_conv_get(PyObject* self, PyObject* args)
+{
+    int sqltype;
+    if (!PyArg_ParseTuple(args, "i", &sqltype))
+        return 0;
+
+    return _get_converter(self, (SQLSMALLINT)sqltype);
+}
+
 static void NormalizeCodecName(const char* src, char* dest, size_t cbDest)
 {
     // Copies the codec name to dest, lowercasing it and replacing underscores with dashes.
@@ -1345,6 +1602,8 @@ static struct PyMethodDef Connection_methods[] =
     { "rollback",                Connection_rollback,        METH_NOARGS,  rollback_doc   },
     { "getinfo",                 Connection_getinfo,         METH_VARARGS, getinfo_doc    },
     { "add_output_converter",    Connection_conv_add,        METH_VARARGS, conv_add_doc   },
+    { "remove_output_converter", Connection_conv_remove,     METH_VARARGS, conv_remove_doc },
+    { "get_output_converter",    Connection_conv_get,        METH_VARARGS, conv_get_doc },
     { "clear_output_converters", Connection_conv_clear,      METH_NOARGS,  conv_clear_doc },
     { "setdecoding",             (PyCFunction)Connection_setdecoding,     METH_VARARGS|METH_KEYWORDS, setdecoding_doc },
     { "setencoding",             (PyCFunction)Connection_setencoding,     METH_VARARGS|METH_KEYWORDS, 0 },
