@@ -24,12 +24,8 @@
 #include "errors.h"
 #include "getdata.h"
 #include "dbspecific.h"
-#include "sqlwchar.h"
 #include <datetime.h>
 
-bool ParamSetup(Cursor *cur, PyObject *sql, PyObject* original_params, bool skip_first);
-int BindAndConvert(Cursor *cur, Py_ssize_t i, PyObject *cell, ParamInfo *ppi);
-
 enum
 {
     CURSOR_REQUIRE_CNXN    = 0x00000001,
@@ -558,118 +554,10 @@ static bool PrepareResults(Cursor* cur, int cCols)
     return true;
 }
 
-bool ProcessDAEParams(SQLRETURN &ret, Cursor *cur, bool freeObj)
-{
-    while (ret == SQL_NEED_DATA)
-    {
-        // One or more parameters were too long to bind normally so we set the
-        // length to SQL_LEN_DATA_AT_EXEC.  ODBC will return SQL_NEED_DATA for
-        // each of the parameters we did this for.
-        //
-        // For each one we set a pointer to the ParamInfo as the "parameter
-        // data" we can access with SQLParamData.  We've stashed everything we
-        // need in there.
-
-        void* pInfo;
-        Py_BEGIN_ALLOW_THREADS
-        ret = SQLParamData(cur->hstmt, (SQLPOINTER*)&pInfo);
-        Py_END_ALLOW_THREADS
-
-        if (ret != SQL_NEED_DATA && ret != SQL_NO_DATA && !SQL_SUCCEEDED(ret))
-            return RaiseErrorFromHandle(cur->cnxn, "SQLParamData", cur->cnxn->hdbc, cur->hstmt);
-
-        TRACE("SQLParamData() --> %d\n", ret);
-        PyObject *obj = ((DAEParam*)pInfo)->cell;
-        SQLLEN maxlen = ((DAEParam*)pInfo)->maxlen;
-
-        if (ret == SQL_NEED_DATA)
-        {
-            if (PyBytes_Check(obj))
-            {
-                const char* p = PyBytes_AS_STRING(obj);
-                SQLLEN offset = 0;
-                SQLLEN cb = (SQLLEN)PyBytes_GET_SIZE(obj);
-                do
-                {
-                    SQLLEN remaining = min(maxlen, cb - offset);
-                    TRACE("SQLPutData [%d] (%d) %.10s\n", offset, remaining, &p[offset]);
-                    Py_BEGIN_ALLOW_THREADS
-                    ret = SQLPutData(cur->hstmt, (SQLPOINTER)&p[offset], remaining);
-                    Py_END_ALLOW_THREADS
-                    if (!SQL_SUCCEEDED(ret))
-                        return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
-                    offset += remaining;
-                }
-                while (offset < cb);
-            }
-            else if (PyUnicode_Check(obj))
-            {
-                const char* p = PyUnicode_AS_DATA(obj);
-                SQLLEN offset = 0;
-                SQLLEN cb = (SQLLEN)PyUnicode_GET_SIZE(obj) * sizeof(SQLWCHAR);
-                do
-                {
-                    SQLLEN remaining = min(maxlen, cb - offset);
-                    TRACE("SQLPutData [%d] (%d) %.10s\n", offset, remaining, &p[offset]);
-                    Py_BEGIN_ALLOW_THREADS
-                    ret = SQLPutData(cur->hstmt, (SQLPOINTER)&p[offset], remaining);
-                    Py_END_ALLOW_THREADS
-                    if (!SQL_SUCCEEDED(ret))
-                        return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
-                    offset += remaining;
-                }
-                while (offset < cb);
-            }
-#if PY_VERSION_HEX >= 0x02060000
-            else if (PyByteArray_Check(obj))
-            {
-                const char* p = PyByteArray_AS_STRING(obj);
-                SQLLEN offset = 0;
-                SQLLEN cb     = (SQLLEN)PyByteArray_GET_SIZE(obj);
-                do
-                {
-                    SQLLEN remaining = min(maxlen, cb - offset);
-                    TRACE("SQLPutData [%d] (%d) %.10s\n", offset, remaining, &p[offset]);
-                    Py_BEGIN_ALLOW_THREADS
-                    ret = SQLPutData(cur->hstmt, (SQLPOINTER)&p[offset], remaining);
-                    Py_END_ALLOW_THREADS
-                    if (!SQL_SUCCEEDED(ret))
-                        return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
-                    offset += remaining;
-                }
-                while (offset < cb);
-            }
-#endif
-#if PY_MAJOR_VERSION < 3
-            else if (PyBuffer_Check(obj))
-            {
-                // Buffers can have multiple segments, so we might need multiple writes.  Looping through buffers isn't
-                // difficult, but we've wrapped it up in an iterator object to keep this loop simple.
-
-                BufferSegmentIterator it(obj);
-                byte* pb;
-                SQLLEN cb;
-                while (it.Next(pb, cb))
-                {
-                    Py_BEGIN_ALLOW_THREADS
-                    ret = SQLPutData(cur->hstmt, pb, cb);
-                    Py_END_ALLOW_THREADS
-                    if (!SQL_SUCCEEDED(ret))
-                        return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
-                }
-            }
-#endif
-            if (freeObj)
-                Py_XDECREF(obj);
-            ret = SQL_NEED_DATA;
-        }
-    }
-    return true;
-}
 
 static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool skip_first)
 {
-    // Internal function to execute SQL, called by .execute and non-fast .executemany.
+    // Internal function to execute SQL, called by .execute and .executemany.
     //
     // pSql
     //   A PyString, PyUnicode, or derived object containing the SQL.
@@ -694,6 +582,7 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
 
     int        params_offset = skip_first ? 1 : 0;
     Py_ssize_t cParams       = params == 0 ? 0 : PySequence_Length(params) - params_offset;
+
     SQLRETURN ret = 0;
 
     free_results(cur, FREE_STATEMENT | KEEP_PREPARED);
@@ -705,16 +594,10 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
         // There are parameters, so we'll need to prepare the SQL statement and bind the parameters.  (We need to
         // prepare the statement because we can't bind a NULL (None) object without knowing the target datatype.  There
         // is no one data type that always maps to the others (no, not even varchar)).
-        if (!ParamSetup(cur, pSql, params, skip_first))
+
+        if (!PrepareAndBind(cur, pSql, params, skip_first))
             return 0;
 
-        // Detect, convert, and bind data
-        for (Py_ssize_t i = 0; i < cParams; i++)
-        {
-            Object param(PySequence_GetItem(params, i + params_offset));
-            if (!BindAndConvert(cur, i, param, &cur->paramInfos[i]))
-                return 0;
-        }
         szLastFunction = "SQLExecute";
         Py_BEGIN_ALLOW_THREADS
         ret = SQLExecute(cur->hstmt);
@@ -777,8 +660,152 @@ static PyObject* execute(Cursor* cur, PyObject* pSql, PyObject* params, bool ski
         return 0;
     }
 
-    if (!ProcessDAEParams(ret, cur, false))
-        return 0;
+    while (ret == SQL_NEED_DATA)
+    {
+        // One or more parameters were too long to bind normally so we set the
+        // length to SQL_LEN_DATA_AT_EXEC.  ODBC will return SQL_NEED_DATA for
+        // each of the parameters we did this for.
+        //
+        // For each one we set a pointer to the ParamInfo as the "parameter
+        // data" we can access with SQLParamData.  We've stashed everything we
+        // need in there.
+
+        szLastFunction = "SQLParamData";
+        ParamInfo* pInfo;
+        Py_BEGIN_ALLOW_THREADS
+        ret = SQLParamData(cur->hstmt, (SQLPOINTER*)&pInfo);
+        Py_END_ALLOW_THREADS
+
+        if (ret != SQL_NEED_DATA && ret != SQL_NO_DATA && !SQL_SUCCEEDED(ret))
+            return RaiseErrorFromHandle(cur->cnxn, "SQLParamData", cur->cnxn->hdbc, cur->hstmt);
+
+        TRACE("SQLParamData() --> %d\n", ret);
+
+        if (ret == SQL_NEED_DATA)
+        {
+            szLastFunction = "SQLPutData";
+            if (pInfo->pObject && (PyBytes_Check(pInfo->pObject)
+    #if PY_VERSION_HEX >= 0x02060000
+             || PyByteArray_Check(pInfo->pObject)
+    #endif
+            ))
+            {
+                char *(*pGetPtr)(PyObject*);
+                Py_ssize_t (*pGetLen)(PyObject*);
+    #if PY_VERSION_HEX >= 0x02060000
+                if (PyByteArray_Check(pInfo->pObject))
+                {
+                    pGetPtr = PyByteArray_AsString;
+                    pGetLen = PyByteArray_Size;
+                }
+                else
+    #endif
+                {
+                    pGetPtr = PyBytes_AsString;
+                    pGetLen = PyBytes_Size;
+                }
+
+                const char* p = pGetPtr(pInfo->pObject);
+                SQLLEN cb = (SQLLEN)pGetLen(pInfo->pObject);
+                SQLLEN offset = 0;
+
+                do
+                {
+                    SQLLEN remaining = pInfo->maxlength ? min(pInfo->maxlength, cb - offset) : cb;
+                    TRACE("SQLPutData [%d] (%d) %.10s\n", offset, remaining, &p[offset]);
+                    Py_BEGIN_ALLOW_THREADS
+                    ret = SQLPutData(cur->hstmt, (SQLPOINTER)&p[offset], remaining);
+                    Py_END_ALLOW_THREADS
+                    if (!SQL_SUCCEEDED(ret))
+                        return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
+                    offset += remaining;
+                }
+                while (offset < cb);
+            }
+#if PY_MAJOR_VERSION < 3
+            else if (pInfo->pObject && PyBuffer_Check(pInfo->pObject))
+            {
+                // Buffers can have multiple segments, so we might need multiple writes.  Looping through buffers isn't
+                // difficult, but we've wrapped it up in an iterator object to keep this loop simple.
+
+                BufferSegmentIterator it(pInfo->pObject);
+                byte* pb;
+                SQLLEN cb;
+                while (it.Next(pb, cb))
+                {
+                    Py_BEGIN_ALLOW_THREADS
+                    ret = SQLPutData(cur->hstmt, pb, cb);
+                    Py_END_ALLOW_THREADS
+                    if (!SQL_SUCCEEDED(ret))
+                        return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
+                }
+            }
+#endif
+            else if (pInfo->ParameterType == SQL_SS_TABLE)
+            {
+                // TVP
+                // Need to convert its columns into the bound row buffers
+                int hasTvpRows = 0;
+                if (pInfo->curTvpRow < PySequence_Length(pInfo->pObject))
+                {
+                    PyObject *tvpRow = PySequence_GetItem(pInfo->pObject, pInfo->curTvpRow);
+                    Py_XDECREF(tvpRow);
+                    for (Py_ssize_t i = 0; i < PySequence_Size(tvpRow); i++)
+                    {
+                        struct ParamInfo newParam;
+                        struct ParamInfo *prevParam = pInfo->nested + i;
+                        PyObject *cell = PySequence_GetItem(tvpRow, i);
+                        Py_XDECREF(cell);
+                        memset(&newParam, 0, sizeof(newParam));
+                        if (!GetParameterInfo(cur, i, cell, newParam, true))
+                        {
+                            // Error converting object
+                            FreeParameterData(cur);
+                            return NULL;
+                        }
+
+                        if((newParam.ValueType != SQL_C_DEFAULT && prevParam->ValueType != SQL_C_DEFAULT) &&
+                           (newParam.ValueType != prevParam->ValueType ||
+                            newParam.ParameterType != prevParam->ParameterType))
+                        {
+                            FreeParameterData(cur);
+                            return RaiseErrorV(0, ProgrammingError, "Type mismatch between TVP row values");
+                        }
+
+                        if (prevParam->allocated)
+                            pyodbc_free(prevParam->ParameterValuePtr);
+                        Py_XDECREF(prevParam->pObject);
+                        newParam.BufferLength = newParam.StrLen_or_Ind;
+                        newParam.StrLen_or_Ind = SQL_DATA_AT_EXEC;
+                        Py_INCREF(cell);
+                        newParam.pObject = cell;
+                        *prevParam = newParam;
+                        if(prevParam->ParameterValuePtr == &newParam.Data)
+                        {
+                            prevParam->ParameterValuePtr = &prevParam->Data;
+                        }
+                    }
+                    pInfo->curTvpRow++;
+                    hasTvpRows = 1;
+                }
+                Py_BEGIN_ALLOW_THREADS
+                ret = SQLPutData(cur->hstmt, hasTvpRows ? (SQLPOINTER)1 : 0, hasTvpRows);
+                Py_END_ALLOW_THREADS
+                if (!SQL_SUCCEEDED(ret))
+                    return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
+            }
+            else
+            {
+                // TVP column sent as DAE
+                Py_BEGIN_ALLOW_THREADS
+                ret = SQLPutData(cur->hstmt, pInfo->ParameterValuePtr, pInfo->BufferLength);
+                Py_END_ALLOW_THREADS
+                if (!SQL_SUCCEEDED(ret))
+                    return RaiseErrorFromHandle(cur->cnxn, "SQLPutData", cur->cnxn->hdbc, cur->hstmt);
+            }
+            ret = SQL_NEED_DATA;
+        }
+    }
 
     FreeParameterData(cur);
 
@@ -943,9 +970,9 @@ static PyObject* Cursor_executemany(PyObject* self, PyObject* args)
         if (cursor->fastexecmany)
         {
             free_results(cursor, FREE_STATEMENT | KEEP_PREPARED);
-            if (!ExecuteMulti(cursor, pSql, param_seq))
+			if (!ExecuteMulti(cursor, pSql, param_seq))
                 return 0;
-        }
+		}
         else
         {
             for (Py_ssize_t i = 0; i < c; i++)
@@ -1002,6 +1029,7 @@ static PyObject* Cursor_executemany(PyObject* self, PyObject* args)
     }
 
     cursor->rowcount = -1;
+
     Py_RETURN_NONE;
 }
 
@@ -1316,12 +1344,12 @@ char* Cursor_column_kwnames[] = { "table", "catalog", "schema", "column", 0 };
 
 static PyObject* Cursor_columns(PyObject* self, PyObject* args, PyObject* kwargs)
 {
-    const char* szCatalog = 0;
-    const char* szSchema  = 0;
-    const char* szTable   = 0;
-    const char* szColumn  = 0;
+    PyObject* pCatalog = 0;
+    PyObject* pSchema  = 0;
+    PyObject* pTable   = 0;
+    PyObject* pColumn  = 0;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|zzzz", Cursor_column_kwnames, &szTable, &szCatalog, &szSchema, &szColumn))
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOO", Cursor_column_kwnames, &pTable, &pCatalog, &pSchema, &pColumn))
         return 0;
 
     Cursor* cur = Cursor_Validate(self, CURSOR_REQUIRE_OPEN);
@@ -1331,8 +1359,21 @@ static PyObject* Cursor_columns(PyObject* self, PyObject* args, PyObject* kwargs
 
     SQLRETURN ret = 0;
 
+    const TextEnc& enc = cur->cnxn->metadata_enc;
+    SQLWChar catalog(pCatalog, enc);
+    SQLWChar schema(pSchema, enc);
+    SQLWChar table(pTable, enc);
+    SQLWChar column(pColumn, enc);
+
+    if (!catalog.isValidOrNone() || !schema.isValidOrNone() || !table.isValidOrNone() || !column.isValidOrNone())
+        return 0;
+
     Py_BEGIN_ALLOW_THREADS
-    ret = SQLColumns(cur->hstmt, (SQLCHAR*)szCatalog, SQL_NTS, (SQLCHAR*)szSchema, SQL_NTS, (SQLCHAR*)szTable, SQL_NTS, (SQLCHAR*)szColumn, SQL_NTS);
+    ret = SQLColumnsW(cur->hstmt,
+                      catalog.psz, SQL_NTS,
+                      schema.psz, SQL_NTS,
+                      table.psz, SQL_NTS,
+                      column.psz, SQL_NTS);
     Py_END_ALLOW_THREADS
 
     if (!SQL_SUCCEEDED(ret))
@@ -2061,11 +2102,6 @@ static char fastexecmany_doc[] =
     "This read/write attribute specifies whether to use a faster executemany() which\n" \
     "uses parameter arrays. Not all drivers may work with this implementation.";
 
-static char decimalasstring_doc[] =
-    "This read/write attribute specifies whether Decimal objects will be bound as strings\n" \
-    "if the server column is of numerical type, for drivers which do not correctly handle\n" \
-    "SQL_NUMERIC_STRUCT.";
-
 static PyMemberDef Cursor_members[] =
 {
     {"rowcount",    T_INT,       offsetof(Cursor, rowcount),        READONLY, rowcount_doc },
@@ -2073,7 +2109,6 @@ static PyMemberDef Cursor_members[] =
     {"arraysize",   T_INT,       offsetof(Cursor, arraysize),       0,        arraysize_doc },
     {"connection",  T_OBJECT_EX, offsetof(Cursor, cnxn),            READONLY, connection_doc },
     {"fast_executemany",T_BOOL,  offsetof(Cursor, fastexecmany),    0,        fastexecmany_doc },
-    {"decimal_as_string",T_BOOL, offsetof(Cursor, decimal_as_string),0,       decimalasstring_doc },
     { 0 }
 };
 
@@ -2168,7 +2203,7 @@ static char fetchone_doc[] =
     "A ProgrammingError exception is raised if the previous call to execute() did\n" \
     "not produce any result set or no call was issued yet.";
 
-static char fetchall_doc[] =
+static char fetchmany_doc[] =
     "fetchmany(size=cursor.arraysize) --> list of Rows\n" \
     "\n" \
     "Fetch the next set of rows of a query result, returning a list of Row\n" \
@@ -2183,8 +2218,8 @@ static char fetchall_doc[] =
     "A ProgrammingError exception is raised if the previous call to execute() did\n" \
     "not produce any result set or no call was issued yet.";
 
-static char fetchmany_doc[] =
-    "fetchmany() --> list of Rows\n" \
+static char fetchall_doc[] =
+    "fetchall() --> list of Rows\n" \
     "\n" \
     "Fetch all remaining rows of a query result, returning them as a list of Rows.\n" \
     "An empty list is returned if there are no more rows.\n" \
@@ -2362,7 +2397,6 @@ Cursor_New(Connection* cnxn)
         cur->rowcount          = -1;
         cur->map_name_to_index = 0;
         cur->fastexecmany      = 0;
-        cur->decimal_as_string = 0;
 
         Py_INCREF(cnxn);
         Py_INCREF(cur->description);
