@@ -56,24 +56,6 @@ typedef unsigned long long UINT64;
 #include <sql.h>
 #include <sqlext.h>
 
-#ifndef SQL_SS_TIMESTAMPOFFSET
-
-#define SQL_SS_TIMESTAMPOFFSET (-155)
-
-typedef struct tagSS_TIMESTAMPOFFSET_STRUCT {  
-    SQLSMALLINT year;  
-    SQLUSMALLINT month;  
-    SQLUSMALLINT day;  
-    SQLUSMALLINT hour;  
-    SQLUSMALLINT minute;  
-    SQLUSMALLINT second;  
-    SQLUINTEGER fraction;  
-    SQLSMALLINT timezone_hour;  
-    SQLSMALLINT timezone_minute;  
-} SQL_SS_TIMESTAMPOFFSET_STRUCT;
-
-#endif
-
 #if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
 typedef int Py_ssize_t;
 #define PY_SSIZE_T_MAX INT_MAX
@@ -88,6 +70,18 @@ typedef int Py_ssize_t;
 #define _countof(a) (sizeof(a) / sizeof(a[0]))
 #endif
 
+#ifndef SQL_SS_TABLE
+#define SQL_SS_TABLE -153
+#endif
+
+#ifndef SQL_SOPT_SS_PARAM_FOCUS
+#define SQL_SOPT_SS_PARAM_FOCUS 1236
+#endif
+
+#ifndef SQL_CA_SS_TYPE_NAME
+#define SQL_CA_SS_TYPE_NAME 1227
+#endif
+
 inline bool IsSet(DWORD grf, DWORD flags)
 {
     return (grf & flags) == flags;
@@ -154,6 +148,11 @@ inline void DebugTrace(const char* szFmt, ...) { UNUSED(szFmt); }
 #define pyodbc_free free
 // #endif
 
+bool pyodbc_realloc(BYTE** pp, size_t newlen);
+// A wrapper around realloc with a safer interface.  If it is successful, *pp is updated to the
+// new pointer value.  If not successful, it is not modified.  (It is easy to forget and lose
+// the old pointer value with realloc.)
+
 void PrintBytes(void* p, size_t len);
 const char* CTypeName(SQLSMALLINT n);
 const char* SqlTypeName(SQLSMALLINT n);
