@@ -126,6 +126,21 @@ static ExcInfo aExcInfos[] = {
 };
 
 
+bool pyodbc_realloc(BYTE** pp, size_t newlen)
+{
+    // A wrapper around realloc with a safer interface.  If it is successful, *pp is updated to the
+    // new pointer value.  If not successful, it is not modified.  (It is easy to forget and lose
+    // the old pointer value with realloc.)
+
+    BYTE* pT = (BYTE*)realloc(*pp, newlen);
+    if (pT == 0)
+        return false;
+    *pp = pT;
+    return true;
+}
+
+
+
 bool UseNativeUUID()
 {
     PyObject* o = PyObject_GetAttrString(pModule, "native_uuid");
@@ -320,6 +335,35 @@ static bool AllocateEnv()
     return true;
 }
 
+static bool CheckAttrsVal(PyObject *val, bool allowSeq)
+{
+    if (IntOrLong_Check(val)
+#if PY_MAJOR_VERSION < 3
+     || PyBuffer_Check(val)
+#endif
+#if PY_VERSION_HEX >= 0x02060000
+     || PyByteArray_Check(val)
+#endif
+     || PyBytes_Check(val)
+     || PyUnicode_Check(val))
+        return true;
+
+    if (allowSeq && PySequence_Check(val))
+    {
+        Py_ssize_t len = PySequence_Size(val);
+        for (Py_ssize_t i = 0; i < len; i++)
+        {
+            Object v(PySequence_GetItem(val, i));
+            if (!CheckAttrsVal(v, false))
+                return false;
+        }
+        return true;
+    }
+
+    return PyErr_Format(PyExc_TypeError, "Attribute dictionary attrs must be"
+        " integers, buffers, bytes, %s", allowSeq ? "strings, or sequences" : "or strings") != 0;
+}
+
 static PyObject* _CheckAttrsDict(PyObject* attrs)
 {
     // The attrs_before dictionary must be keys to integer values.  If valid and non-empty,
@@ -340,8 +384,9 @@ static PyObject* _CheckAttrsDict(PyObject* attrs)
     {
         if (!IntOrLong_Check(key))
             return PyErr_Format(PyExc_TypeError, "Attribute dictionary keys must be integers");
-        if (!IntOrLong_Check(value))
-            return PyErr_Format(PyExc_TypeError, "Attribute dictionary attrs must be integers");
+
+        if (!CheckAttrsVal(value, true))
+            return 0;
     }
     Py_INCREF(attrs);
     return attrs;
@@ -441,7 +486,7 @@ static PyObject* mod_connect(PyObject* self, PyObject* args, PyObject* kwargs)
                 fReadOnly = PyObject_IsTrue(value);
                 continue;
             }
-            if (Text_EqualsI(key, "attrs_before"))
+            if (Text_EqualsI(key, "attrs_before") && PyDict_Check(value))
             {
                 attrs_before = _CheckAttrsDict(value);
                 if (PyErr_Occurred())
@@ -506,7 +551,7 @@ static PyObject* mod_connect(PyObject* self, PyObject* args, PyObject* kwargs)
     }
 
     return (PyObject*)Connection_New(pConnectString.Get(), fAutoCommit != 0, fAnsi != 0, timeout,
-                                     fReadOnly != 0, attrs_before, encoding);
+                                     fReadOnly != 0, attrs_before.Detach(), encoding);
 }
 
 
@@ -904,7 +949,6 @@ static const ConstantDef aConstants[] = {
     MAKECONST(SQL_COLLATION_SEQ),
     MAKECONST(SQL_COLUMN_ALIAS),
     MAKECONST(SQL_CONCAT_NULL_BEHAVIOR),
-    MAKECONST(SQL_CONVERT_FUNCTIONS),
     MAKECONST(SQL_CONVERT_VARCHAR),
     MAKECONST(SQL_CORRELATION_NAME),
     MAKECONST(SQL_CREATE_ASSERTION),
@@ -1039,14 +1083,57 @@ static const ConstantDef aConstants[] = {
     MAKECONST(SQL_UNION),
     MAKECONST(SQL_USER_NAME),
     MAKECONST(SQL_XOPEN_CLI_YEAR),
-    
+
+    // Connection Attributes
+    MAKECONST(SQL_ACCESS_MODE), MAKECONST(SQL_ATTR_ACCESS_MODE),
+    MAKECONST(SQL_AUTOCOMMIT), MAKECONST(SQL_ATTR_AUTOCOMMIT),
+    MAKECONST(SQL_LOGIN_TIMEOUT), MAKECONST(SQL_ATTR_LOGIN_TIMEOUT),
+    MAKECONST(SQL_OPT_TRACE), MAKECONST(SQL_ATTR_TRACE),
+    MAKECONST(SQL_OPT_TRACEFILE), MAKECONST(SQL_ATTR_TRACEFILE),
+    MAKECONST(SQL_TRANSLATE_DLL), MAKECONST(SQL_ATTR_TRANSLATE_LIB),
+    MAKECONST(SQL_TRANSLATE_OPTION), MAKECONST(SQL_ATTR_TRANSLATE_OPTION),
+    MAKECONST(SQL_TXN_ISOLATION), MAKECONST(SQL_ATTR_TXN_ISOLATION),
+    MAKECONST(SQL_CURRENT_QUALIFIER), MAKECONST(SQL_ATTR_CURRENT_CATALOG),
+    MAKECONST(SQL_ODBC_CURSORS), MAKECONST(SQL_ATTR_ODBC_CURSORS),
+    MAKECONST(SQL_QUIET_MODE), MAKECONST(SQL_ATTR_QUIET_MODE),
+    MAKECONST(SQL_PACKET_SIZE),
+    MAKECONST(SQL_ATTR_ANSI_APP),
+
+    // SQL_CONVERT_X
+    MAKECONST(SQL_CONVERT_FUNCTIONS),
+    MAKECONST(SQL_CONVERT_BIGINT),
+    MAKECONST(SQL_CONVERT_BINARY),
+    MAKECONST(SQL_CONVERT_BIT),
+    MAKECONST(SQL_CONVERT_CHAR),
+    MAKECONST(SQL_CONVERT_DATE),
+    MAKECONST(SQL_CONVERT_DECIMAL),
+    MAKECONST(SQL_CONVERT_DOUBLE),
+    MAKECONST(SQL_CONVERT_FLOAT),
+    MAKECONST(SQL_CONVERT_GUID),
+    MAKECONST(SQL_CONVERT_INTEGER),
+    MAKECONST(SQL_CONVERT_INTERVAL_DAY_TIME),
+    MAKECONST(SQL_CONVERT_INTERVAL_YEAR_MONTH),
+    MAKECONST(SQL_CONVERT_LONGVARBINARY),
+    MAKECONST(SQL_CONVERT_LONGVARCHAR),
+    MAKECONST(SQL_CONVERT_NUMERIC),
+    MAKECONST(SQL_CONVERT_REAL),
+    MAKECONST(SQL_CONVERT_SMALLINT),
+    MAKECONST(SQL_CONVERT_TIME),
+    MAKECONST(SQL_CONVERT_TIMESTAMP),
+    MAKECONST(SQL_CONVERT_TINYINT),
+    MAKECONST(SQL_CONVERT_VARBINARY),
+    MAKECONST(SQL_CONVERT_VARCHAR),
+    MAKECONST(SQL_CONVERT_WCHAR),
+    MAKECONST(SQL_CONVERT_WLONGVARCHAR),
+    MAKECONST(SQL_CONVERT_WVARCHAR),
+
     // SQLSetConnectAttr transaction isolation
     MAKECONST(SQL_ATTR_TXN_ISOLATION),
     MAKECONST(SQL_TXN_READ_UNCOMMITTED),
     MAKECONST(SQL_TXN_READ_COMMITTED),
     MAKECONST(SQL_TXN_REPEATABLE_READ),
     MAKECONST(SQL_TXN_SERIALIZABLE),
-    
+
     // Outer Join Capabilities
     MAKECONST(SQL_OJ_LEFT),
     MAKECONST(SQL_OJ_RIGHT),
