@@ -9,25 +9,25 @@ These run using the version from the 'build' directory, not the version
 installed into the Python directories.  You must run python setup.py build
 before running the tests.
 
-To run, pass the filename of an Access database on the command line:
+To run, pass the file EXTENSION of an Access database on the command line:
 
-  accesstests test.accdb
+  accesstests accdb
 
-An empty Access 2000 database (empty.mdb) and an empty Access 2007 database
-(empty.accdb), are provided.
+An empty Access 2000 database (empty.mdb) or an empty Access 2007 database
+(empty.accdb), are automatically created for the tests.
 
 To run a single test, use the -t option:
 
-  accesstests test.accdb -t unicode_null
+  accesstests -t unicode_null accdb
 
 If you want to report an error, it would be helpful to include the driver information
 by using the verbose flag and redirecting the output to a file:
 
- accesstests test.accdb -v >& results.txt
+ accesstests -v accdb >& results.txt
 
 You can pass the verbose flag twice for more verbose output:
 
- accesstests test.accdb -vv
+ accesstests -vv accdb
 """
 
 # Access SQL data types: http://msdn2.microsoft.com/en-us/library/bb208866.aspx
@@ -36,7 +36,8 @@ import sys, os, re
 import unittest
 from decimal import Decimal
 from datetime import datetime, date, time
-from os.path import abspath
+from os.path import abspath, dirname, join
+import shutil
 from testutils import *
 
 CNXNSTRING = None
@@ -77,6 +78,11 @@ class AccessTestCase(unittest.TestCase):
         self.cnxn   = pyodbc.connect(CNXNSTRING)
         self.cursor = self.cnxn.cursor()
 
+        # https://docs.microsoft.com/en-us/sql/odbc/microsoft/desktop-database-driver-performance-issues?view=sql-server-2017
+        #
+        # As of the 4.0 drivers, you have to send as Unicode?
+        self.cnxn.setencoding(str, encoding='utf-16le')
+
         for i in range(3):
             try:
                 self.cursor.execute("drop table t%d" % i)
@@ -113,27 +119,27 @@ class AccessTestCase(unittest.TestCase):
 
     def test_drivers(self):
         p = pyodbc.drivers()
-        self.assert_(isinstance(p, list))
+        self.assertTrue(isinstance(p, list))
 
     def test_datasources(self):
         p = pyodbc.dataSources()
-        self.assert_(isinstance(p, dict))
+        self.assertTrue(isinstance(p, dict))
 
     def test_getinfo_string(self):
         value = self.cnxn.getinfo(pyodbc.SQL_CATALOG_NAME_SEPARATOR)
-        self.assert_(isinstance(value, str))
+        self.assertTrue(isinstance(value, str))
 
     def test_getinfo_bool(self):
         value = self.cnxn.getinfo(pyodbc.SQL_ACCESSIBLE_TABLES)
-        self.assert_(isinstance(value, bool))
+        self.assertTrue(isinstance(value, bool))
 
     def test_getinfo_int(self):
         value = self.cnxn.getinfo(pyodbc.SQL_DEFAULT_TXN_ISOLATION)
-        self.assert_(isinstance(value, (int, long)))
+        self.assertTrue(isinstance(value, (int, long)))
 
     def test_getinfo_smallint(self):
         value = self.cnxn.getinfo(pyodbc.SQL_CONCAT_NULL_BEHAVIOR)
-        self.assert_(isinstance(value, int))
+        self.assertTrue(isinstance(value, int))
 
     def _test_strtype(self, sqltype, value, resulttype=None, colsize=None):
         """
@@ -297,11 +303,11 @@ class AccessTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(s varchar(20))")
         self.cursor.execute("insert into t1 values(?)", "1")
         row = self.cursor.execute("select * from t1").fetchone()
-        self.assertEquals(row[0], "1")
-        self.assertEquals(row[-1], "1")
+        self.assertEqual(row[0], "1")
+        self.assertEqual(row[-1], "1")
 
     def test_version(self):
-        self.assertEquals(3, len(pyodbc.version.split('.'))) # 1.3.1 etc.
+        self.assertEqual(3, len(pyodbc.version.split('.'))) # 1.3.1 etc.
 
     #
     # date, time, datetime
@@ -314,7 +320,7 @@ class AccessTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", value)
 
         result = self.cursor.execute("select dt from t1").fetchone()[0]
-        self.assertEquals(value, result)
+        self.assertEqual(value, result)
 
     #
     # ints and floats
@@ -325,28 +331,28 @@ class AccessTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(n int)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_negative_int(self):
         value = -1
         self.cursor.execute("create table t1(n int)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_smallint(self):
         value = 32767
         self.cursor.execute("create table t1(n smallint)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_real(self):
         value = 1234.5
         self.cursor.execute("create table t1(n real)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_negative_real(self):
         value = -200.5
@@ -360,7 +366,7 @@ class AccessTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(n float)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_negative_float(self):
         value = -200.5
@@ -428,10 +434,7 @@ class AccessTestCase(unittest.TestCase):
         self.assertEqual(False, result)
 
     def test_guid(self):
-        # REVIEW: Python doesn't (yet) have a UUID type so the value is returned as a string.  Access, however, only
-        # really supports Unicode.  For now, we'll have to live with this difference.  All strings in Python 3.x will
-        # be Unicode -- pyodbc 3.x will have different defaults.
-        value = "de2ac9c6-8676-4b0b-b8a6-217a8580cbee"
+        value = u"de2ac9c6-8676-4b0b-b8a6-217a8580cbee"
         self.cursor.execute("create table t1(g1 uniqueidentifier)")
         self.cursor.execute("insert into t1 values (?)", value)
         v = self.cursor.execute("select * from t1").fetchone()[0]
@@ -444,13 +447,13 @@ class AccessTestCase(unittest.TestCase):
     #
 
     def test_rowcount_delete(self):
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, -1)
         self.cursor.execute("create table t1(i int)")
         count = 4
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
         self.cursor.execute("delete from t1")
-        self.assertEquals(self.cursor.rowcount, count)
+        self.assertEqual(self.cursor.rowcount, count)
 
     def test_rowcount_nodata(self):
         """
@@ -463,7 +466,7 @@ class AccessTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(i int)")
         # This is a different code path internally.
         self.cursor.execute("delete from t1")
-        self.assertEquals(self.cursor.rowcount, 0)
+        self.assertEqual(self.cursor.rowcount, 0)
 
     def test_rowcount_select(self):
         """
@@ -478,11 +481,11 @@ class AccessTestCase(unittest.TestCase):
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
         self.cursor.execute("select * from t1")
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, -1)
 
         rows = self.cursor.fetchall()
-        self.assertEquals(len(rows), count)
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(len(rows), count)
+        self.assertEqual(self.cursor.rowcount, -1)
 
     def test_rowcount_reset(self):
         "Ensure rowcount is reset to -1"
@@ -491,10 +494,10 @@ class AccessTestCase(unittest.TestCase):
         count = 4
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
-        self.assertEquals(self.cursor.rowcount, 1)
+        self.assertEqual(self.cursor.rowcount, 1)
 
         self.cursor.execute("create table t2(i int)")
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, -1)
 
     #
     # Misc
@@ -514,7 +517,7 @@ class AccessTestCase(unittest.TestCase):
         names = [ t[0] for t in self.cursor.description ]
         names.sort()
 
-        self.assertEquals(names, [ "abc", "def" ])
+        self.assertEqual(names, [ "abc", "def" ])
 
         # Put it back so other tests don't fail.
         pyodbc.lowercase = False
@@ -529,7 +532,7 @@ class AccessTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values(1, 'abc')")
 
         row = self.cursor.execute("select * from t1").fetchone()
-        self.assertEquals(self.cursor.description, row.cursor_description)
+        self.assertEqual(self.cursor.description, row.cursor_description)
         
 
     def test_executemany(self):
@@ -561,7 +564,7 @@ class AccessTestCase(unittest.TestCase):
                    ('error', 'not an int'),
                    (3, 'good') ]
         
-        self.failUnlessRaises(pyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)
+        self.assertRaises(pyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)
 
         
     def test_row_slicing(self):
@@ -571,13 +574,13 @@ class AccessTestCase(unittest.TestCase):
         row = self.cursor.execute("select * from t1").fetchone()
 
         result = row[:]
-        self.failUnless(result is row)
+        self.assertTrue(result is row)
 
         result = row[:-1]
         self.assertEqual(result, (1,2,3))
 
         result = row[0:4]
-        self.failUnless(result is row)
+        self.assertTrue(result is row)
 
 
     def test_row_repr(self):
@@ -622,7 +625,7 @@ class AccessTestCase(unittest.TestCase):
 def main():
     from optparse import OptionParser
     parser = OptionParser(usage=usage)
-    parser.add_option("-v", "--verbose", action="count", help="Increment test verbosity (can be used multiple times)")
+    parser.add_option("-v", "--verbose", default=0, action="count", help="Increment test verbosity (can be used multiple times)")
     parser.add_option("-d", "--debug", action="store_true", default=False, help="Print debugging items")
     parser.add_option("-t", "--test", help="Run only the named test")
 
@@ -633,11 +636,19 @@ def main():
 
     if args[0].endswith('.accdb'):
         driver = 'Microsoft Access Driver (*.mdb, *.accdb)'
+        drvext = 'accdb'
     else:
         driver = 'Microsoft Access Driver (*.mdb)'
+        drvext = 'mdb'
+
+    here = dirname(abspath(__file__))
+    src = join(here, 'empty.' + drvext)
+    dest = join(here, 'test.' + drvext)
+    shutil.copy(src, dest)
 
     global CNXNSTRING
-    CNXNSTRING = 'DRIVER={%s};DBQ=%s;ExtendedAnsiSQL=1' % (driver, abspath(args[0]))
+    CNXNSTRING = 'DRIVER={%s};DBQ=%s;ExtendedAnsiSQL=1' % (driver, dest)
+    print(CNXNSTRING)
 
     cnxn = pyodbc.connect(CNXNSTRING)
     print_library_info(cnxn)
@@ -648,10 +659,12 @@ def main():
     testRunner = unittest.TextTestRunner(verbosity=options.verbose)
     result = testRunner.run(suite)
 
+    return result
+
 
 if __name__ == '__main__':
 
     # Add the build directory to the path so we're testing the latest build, not the installed version.
     add_to_path()
     import pyodbc
-    main()
+    sys.exit(0 if main().wasSuccessful() else 1)
