@@ -13,8 +13,7 @@ These run using the version from the 'build' directory, not the version
 installed into the Python directories.  You must run python setup.py build
 before running the tests.
 
-You can also put the connection string into a setup.cfg file in the root of the project
-(the same one setup.py would use) like so:
+You can also put the connection string into a tmp/setup.cfg file like so:
 
   [sqlservertests]
   connection-string=DRIVER={SQL Server};SERVER=localhost;UID=uid;PWD=pwd;DATABASE=db
@@ -25,6 +24,8 @@ is installed:
   2000: DRIVER={SQL Server}
   2005: DRIVER={SQL Server}
   2008: DRIVER={SQL Server Native Client 10.0}
+  
+If using FreeTDS ODBC, be sure to use version 1.00.97 or newer.
 """
 
 import sys, os, re, uuid
@@ -32,8 +33,15 @@ import unittest
 from decimal import Decimal
 from datetime import datetime, date, time
 from os.path import join, getsize, dirname, abspath
+from warnings import warn
 from testutils import *
 
+# Some tests have fallback code for known driver issues.
+# Change this value to False to bypass the fallback code, e.g., to see
+#   if a newer version of the driver has fixed the underlying issue.
+#
+handle_known_issues = True
+
 _TESTSTR = '0123456789-abcdefghijklmnopqrstuvwxyz-'
 
 def _generate_test_string(length):
@@ -72,6 +80,67 @@ class SqlServerTestCase(unittest.TestCase):
         unittest.TestCase.__init__(self, method_name)
         self.connection_string = connection_string
 
+    def driver_type_is(self, type_name):
+        recognized_types = {
+            'msodbcsql': '(Microsoft) ODBC Driver xx for SQL Server',
+            'freetds': 'FreeTDS ODBC',
+        }
+        if not type_name in recognized_types.keys():
+            raise KeyError('"{0}" is not a recognized driver type: {1}'.format(type_name, list(recognized_types.keys())))
+        driver_name = self.cnxn.getinfo(pyodbc.SQL_DRIVER_NAME).lower()
+        if type_name == 'msodbcsql':
+            return ('msodbcsql' in driver_name) or ('sqlncli' in driver_name) or ('sqlsrv32.dll' == driver_name)
+        elif type_name == 'freetds':
+            return ('tdsodbc' in driver_name)
+
+    def handle_known_issues_for(self, type_name, print_reminder=False):
+        """
+        Checks driver `type_name` and "killswitch" variable `handle_known_issues` to see if
+        known issue handling should be bypassed. Optionally prints a reminder message to
+        help identify tests that previously had issues but may have been fixed by a newer
+        version of the driver.
+
+        Usage examples:
+
+        # 1. print reminder at beginning of test (before any errors can occur)
+        #
+        def test_some_feature(self):
+            self.handle_known_issues_for('freetds', print_reminder=True)
+            # (continue with test code)
+
+        # 2. conditional execution of fallback code
+        #
+        try:
+            # (some test code)
+        except pyodbc.DataError:
+            if self.handle_known_issues_for('freetds'):
+                # FREETDS_KNOWN_ISSUE
+                #
+                # (fallback code to work around exception)
+            else:
+                raise
+        """
+        if self.driver_type_is(type_name):
+            if handle_known_issues:
+                return True
+            else:
+                if print_reminder:
+                    print("Known issue handling is disabled. Does this test still fail?")
+        return False
+
+    def driver_type_is(self, type_name):
+        recognized_types = {
+            'msodbcsql': '(Microsoft) ODBC Driver xx for SQL Server',
+            'freetds': 'FreeTDS ODBC',
+        }
+        if not type_name in recognized_types.keys():
+            raise KeyError('"{0}" is not a recognized driver type: {1}'.format(type_name, list(recognized_types.keys())))
+        driver_name = self.cnxn.getinfo(pyodbc.SQL_DRIVER_NAME).lower()
+        if type_name == 'msodbcsql':
+            return ('msodbcsql' in driver_name) or ('sqlncli' in driver_name) or ('sqlsrv32.dll' == driver_name)
+        elif type_name == 'freetds':
+            return ('tdsodbc' in driver_name)
+
     def get_sqlserver_version(self):
         """
         Returns the major version: 8-->2000, 9-->2005, 10-->2008
@@ -116,9 +185,9 @@ class SqlServerTestCase(unittest.TestCase):
 
     def test_binary_type(self):
         if sys.hexversion >= 0x02060000:
-            self.assertIs(pyodbc.BINARY, bytearray)
+            self.assertTrue(pyodbc.BINARY is bytearray)
         else:
-            self.assertIs(pyodbc.BINARY, buffer)
+            self.assertTrue(pyodbc.BINARY is buffer)
 
     def test_multiple_bindings(self):
         "More than one bind and select on a cursor"
@@ -139,27 +208,27 @@ class SqlServerTestCase(unittest.TestCase):
 
     def test_drivers(self):
         p = pyodbc.drivers()
-        self.assert_(isinstance(p, list))
+        self.assertTrue(isinstance(p, list))
 
     def test_datasources(self):
         p = pyodbc.dataSources()
-        self.assert_(isinstance(p, dict))
+        self.assertTrue(isinstance(p, dict))
 
     def test_getinfo_string(self):
         value = self.cnxn.getinfo(pyodbc.SQL_CATALOG_NAME_SEPARATOR)
-        self.assert_(isinstance(value, str))
+        self.assertTrue(isinstance(value, str))
 
     def test_getinfo_bool(self):
         value = self.cnxn.getinfo(pyodbc.SQL_ACCESSIBLE_TABLES)
-        self.assert_(isinstance(value, bool))
+        self.assertTrue(isinstance(value, bool))
 
     def test_getinfo_int(self):
         value = self.cnxn.getinfo(pyodbc.SQL_DEFAULT_TXN_ISOLATION)
-        self.assert_(isinstance(value, (int, long)))
+        self.assertTrue(isinstance(value, (int, long)))
 
     def test_getinfo_smallint(self):
         value = self.cnxn.getinfo(pyodbc.SQL_CONCAT_NULL_BEHAVIOR)
-        self.assert_(isinstance(value, int))
+        self.assertTrue(isinstance(value, int))
 
     def test_noscan(self):
         self.assertEqual(self.cursor.noscan, False)
@@ -186,7 +255,7 @@ class SqlServerTestCase(unittest.TestCase):
 
         pyodbc.native_uuid = True
         result = self.cursor.execute("select n from t1").fetchval()
-        self.assertIsInstance(result, uuid.UUID)
+        self.assertTrue(isinstance(result, uuid.UUID))
         self.assertEqual(value, result)
 
     def test_nextset(self):
@@ -208,6 +277,11 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("select i = 1; RAISERROR('c', 16, 1);")
         row = next(self.cursor)
         self.assertEqual(1, row.i)
+        if self.driver_type_is('freetds'):
+            warn('FREETDS_KNOWN_ISSUE - test_nextset_with_raiserror: test cancelled.')
+            # AssertionError: ProgrammingError not raised by nextset
+            # https://github.com/FreeTDS/freetds/issues/230
+            return  # for now
         self.assertRaises(pyodbc.ProgrammingError, self.cursor.nextset)
 
     def test_fixed_unicode(self):
@@ -231,12 +305,32 @@ class SqlServerTestCase(unittest.TestCase):
             sql = "create table t1(s %s(%s))" % (sqltype, colsize)
         else:
             sql = "create table t1(s %s)" % sqltype
+        self.cursor.execute(sql)
 
         if resulttype is None:
             resulttype = type(value)
 
-        self.cursor.execute(sql)
-        self.cursor.execute("insert into t1 values(?)", value)
+        sql = "insert into t1 values(?)"
+        try:
+            self.cursor.execute(sql, value)
+        except pyodbc.DataError:
+            if self.driver_type_is('freetds'):
+                # FREETDS_KNOWN_ISSUE
+                #
+                # cnxn.getinfo(pyodbc.SQL_DESCRIBE_PARAMETER) returns False for FreeTDS, so
+                # pyodbc can't call SQLDescribeParam to get the correct parameter type.
+                # This can lead to errors being returned from SQL Server when sp_prepexec is called, 
+                # e.g., "Implicit conversion from data type varchar to varbinary is not allowed." 
+                # for test_binary_null
+                #
+                # So at least verify that the user can manually specify the parameter type
+                if sqltype == 'varbinary':
+                    sql_param_type = pyodbc.SQL_VARBINARY
+                    # (add elif blocks for other cases as required)
+                self.cursor.setinputsizes([(sql_param_type, colsize, 0)])
+                self.cursor.execute(sql, value)
+            else:
+                raise
         v = self.cursor.execute("select * from t1").fetchone()[0]
 
         # To allow buffer --> db --> bytearray tests, always convert the input to the expected result type before
@@ -351,6 +445,41 @@ class SqlServerTestCase(unittest.TestCase):
             return              # so pass / ignore
         self.cursor.execute("select cast(replicate(N'x', 512) as nvarchar(max))")
 
+    def test_fast_executemany_to_local_temp_table(self):
+        if self.driver_type_is('freetds'):
+            warn('FREETDS_KNOWN_ISSUE - test_fast_executemany_to_local_temp_table: test cancelled.')
+            return 
+        v = u'Ώπα'
+        self.cursor.execute("CREATE TABLE #issue295 (id INT IDENTITY PRIMARY KEY, txt NVARCHAR(50))")
+        sql = "INSERT INTO #issue295 (txt) VALUES (?)"
+        params = [(v,)]
+        self.cursor.setinputsizes([(pyodbc.SQL_WVARCHAR, 50, 0)])
+        self.cursor.fast_executemany = True
+        self.cursor.executemany(sql, params)
+        self.assertEqual(self.cursor.execute("SELECT txt FROM #issue295").fetchval(), v)
+
+    def test_fast_executemany_to_datetime2(self):
+        if self.handle_known_issues_for('freetds', print_reminder=True):
+            warn('FREETDS_KNOWN_ISSUE - test_fast_executemany_to_datetime2: test cancelled.')
+            return
+        v = datetime(2019, 3, 12, 10, 0, 0, 123456)
+        self.cursor.execute("CREATE TABLE ##issue540 (dt2 DATETIME2(2))")
+        sql = "INSERT INTO ##issue540 (dt2) VALUES (?)"
+        params = [(v,)]
+        self.cursor.fast_executemany = True
+        self.cursor.executemany(sql, params)
+        self.assertEqual(self.cursor.execute("SELECT CAST(dt2 AS VARCHAR) FROM ##issue540").fetchval(), '2019-03-12 10:00:00.12')
+
+    def test_fast_executemany_high_unicode(self):
+        if self.handle_known_issues_for('freetds', print_reminder=True):
+            warn('FREETDS_KNOWN_ISSUE - test_fast_executemany_high_unicode: test cancelled.')
+            return
+        v = u"🎥"
+        self.cursor.fast_executemany = True
+        self.cursor.execute("CREATE TABLE t1 (col1 nvarchar(max) null)")
+        self.cursor.executemany("INSERT INTO t1 (col1) VALUES (?)", [[v,]])
+        self.assertEqual(self.cursor.execute("SELECT * FROM t1").fetchone()[0], v)
+
     #
     # binary
     #
@@ -648,11 +777,11 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(s varchar(20))")
         self.cursor.execute("insert into t1 values(?)", "1")
         row = self.cursor.execute("select * from t1").fetchone()
-        self.assertEquals(row[0], "1")
-        self.assertEquals(row[-1], "1")
+        self.assertEqual(row[0], "1")
+        self.assertEqual(row[-1], "1")
 
     def test_version(self):
-        self.assertEquals(3, len(pyodbc.version.split('.'))) # 1.3.1 etc.
+        self.assertEqual(3, len(pyodbc.version.split('.'))) # 1.3.1 etc.
 
     #
     # date, time, datetime
@@ -665,8 +794,8 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", value)
 
         result = self.cursor.execute("select dt from t1").fetchone()[0]
-        self.assertEquals(type(value), datetime)
-        self.assertEquals(value, result)
+        self.assertEqual(type(result), datetime)
+        self.assertEqual(value, result)
 
     def test_datetime_fraction(self):
         # SQL Server supports milliseconds, but Python's datetime supports nanoseconds, so the most granular datetime
@@ -678,8 +807,8 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", value)
 
         result = self.cursor.execute("select dt from t1").fetchone()[0]
-        self.assertEquals(type(value), datetime)
-        self.assertEquals(result, value)
+        self.assertEqual(type(result), datetime)
+        self.assertEqual(value, result)
 
     def test_datetime_fraction_rounded(self):
         # SQL Server supports milliseconds, but Python's datetime supports nanoseconds.  pyodbc rounds down to what the
@@ -692,8 +821,8 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", full)
 
         result = self.cursor.execute("select dt from t1").fetchone()[0]
-        self.assertEquals(type(result), datetime)
-        self.assertEquals(result, rounded)
+        self.assertEqual(type(result), datetime)
+        self.assertEqual(rounded, result)
 
     def test_date(self):
         ver = self.get_sqlserver_version()
@@ -706,8 +835,8 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", value)
 
         result = self.cursor.execute("select d from t1").fetchone()[0]
-        self.assertEquals(type(value), date)
-        self.assertEquals(value, result)
+        self.assertEqual(type(result), date)
+        self.assertEqual(value, result)
 
     def test_time(self):
         ver = self.get_sqlserver_version()
@@ -724,8 +853,8 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", value)
 
         result = self.cursor.execute("select t from t1").fetchone()[0]
-        self.assertEquals(type(value), time)
-        self.assertEquals(value, result)
+        self.assertEqual(type(result), time)
+        self.assertEqual(value, result)
 
     def test_datetime2(self):
         value = datetime(2007, 1, 15, 3, 4, 5)
@@ -734,8 +863,8 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (?)", value)
 
         result = self.cursor.execute("select dt from t1").fetchone()[0]
-        self.assertEquals(type(value), datetime)
-        self.assertEquals(value, result)
+        self.assertEqual(type(result), datetime)
+        self.assertEqual(value, result)
 
     #
     # ints and floats
@@ -746,14 +875,14 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(n int)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_negative_int(self):
         value = -1
         self.cursor.execute("create table t1(n int)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
 
     def test_bigint(self):
         input = 3000000000
@@ -762,12 +891,28 @@ class SqlServerTestCase(unittest.TestCase):
         result = self.cursor.execute("select d from t1").fetchone()[0]
         self.assertEqual(result, input)
 
+    def test_overflow_int(self):
+        # python allows integers of any size, bigger than an 8 byte int can contain
+        input = 9999999999999999999999999999999999999
+        self.cursor.execute("create table t1(d bigint)")
+        self.cnxn.commit()
+        self.assertRaises(OverflowError, self.cursor.execute, "insert into t1 values (?)", input)
+        result = self.cursor.execute("select * from t1").fetchall()
+        self.assertEqual(result, [])
+
     def test_float(self):
         value = 1234.567
         self.cursor.execute("create table t1(n float)")
         self.cursor.execute("insert into t1 values (?)", value)
         result = self.cursor.execute("select n from t1").fetchone()[0]
-        self.assertEquals(result, value)
+        self.assertEqual(result, value)
+
+    def test_denorm_float(self):
+        value = 0.00012345
+        self.cursor.execute("create table t1(n float)")
+        self.cursor.execute("insert into t1 values (?)", value)
+        result = self.cursor.execute("select n from t1").fetchone()[0]
+        self.assertEqual(result, value)
 
     def test_negative_float(self):
         value = -200
@@ -776,6 +921,14 @@ class SqlServerTestCase(unittest.TestCase):
         result  = self.cursor.execute("select n from t1").fetchone()[0]
         self.assertEqual(value, result)
 
+    def test_non_numeric_float(self):
+        self.cursor.execute("create table t1(d float)")
+        self.cnxn.commit()
+        for input in (float('+Infinity'), float('-Infinity'), float('NaN')):
+            self.assertRaises(pyodbc.ProgrammingError, self.cursor.execute, "insert into t1 values (?)", input)
+        result = self.cursor.execute("select * from t1").fetchall()
+        self.assertEqual(result, [])
+
 
     #
     # stored procedures
@@ -794,9 +947,9 @@ class SqlServerTestCase(unittest.TestCase):
               from sysobjects
             """)
         rows = self.cursor.execute("exec proc1").fetchall()
-        self.assertEquals(type(rows), list)
-        self.assertEquals(len(rows), 10) # there has to be at least 10 items in sysobjects
-        self.assertEquals(type(rows[0].refdate), datetime)
+        self.assertEqual(type(rows), list)
+        self.assertEqual(len(rows), 10) # there has to be at least 10 items in sysobjects
+        self.assertEqual(type(rows[0].refdate), datetime)
 
 
     def test_sp_results_from_temp(self):
@@ -816,13 +969,13 @@ class SqlServerTestCase(unittest.TestCase):
               select * from #tmptable
             """)
         self.cursor.execute("exec proc1")
-        self.assert_(self.cursor.description is not None)
-        self.assert_(len(self.cursor.description) == 4)
+        self.assertTrue(self.cursor.description is not None)
+        self.assertTrue(len(self.cursor.description) == 4)
 
         rows = self.cursor.fetchall()
-        self.assertEquals(type(rows), list)
-        self.assertEquals(len(rows), 10) # there has to be at least 10 items in sysobjects
-        self.assertEquals(type(rows[0].refdate), datetime)
+        self.assertEqual(type(rows), list)
+        self.assertEqual(len(rows), 10) # there has to be at least 10 items in sysobjects
+        self.assertEqual(type(rows[0].refdate), datetime)
 
 
     def test_sp_results_from_vartbl(self):
@@ -841,9 +994,9 @@ class SqlServerTestCase(unittest.TestCase):
             """)
         self.cursor.execute("exec proc1")
         rows = self.cursor.fetchall()
-        self.assertEquals(type(rows), list)
-        self.assertEquals(len(rows), 10) # there has to be at least 10 items in sysobjects
-        self.assertEquals(type(rows[0].refdate), datetime)
+        self.assertEqual(type(rows), list)
+        self.assertEqual(len(rows), 10) # there has to be at least 10 items in sysobjects
+        self.assertEqual(type(rows[0].refdate), datetime)
 
     def test_sp_with_dates(self):
         # Reported in the forums that passing two datetimes to a stored procedure doesn't work.
@@ -862,8 +1015,8 @@ class SqlServerTestCase(unittest.TestCase):
             """)
         self.cursor.execute("exec test_sp ?, ?", datetime.now(), datetime.now())
         rows = self.cursor.fetchall()
-        self.assert_(rows is not None)
-        self.assert_(rows[0][0] == 0)   # 0 years apart
+        self.assertTrue(rows is not None)
+        self.assertTrue(rows[0][0] == 0)   # 0 years apart
 
     def test_sp_with_none(self):
         # Reported in the forums that passing None caused an error.
@@ -882,8 +1035,8 @@ class SqlServerTestCase(unittest.TestCase):
             """)
         self.cursor.execute("exec test_sp ?", None)
         rows = self.cursor.fetchall()
-        self.assert_(rows is not None)
-        self.assert_(rows[0][0] == None)   # 0 years apart
+        self.assertTrue(rows is not None)
+        self.assertTrue(rows[0][0] == None)   # 0 years apart
 
 
     #
@@ -891,13 +1044,13 @@ class SqlServerTestCase(unittest.TestCase):
     #
 
     def test_rowcount_delete(self):
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, -1)
         self.cursor.execute("create table t1(i int)")
         count = 4
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
         self.cursor.execute("delete from t1")
-        self.assertEquals(self.cursor.rowcount, count)
+        self.assertEqual(self.cursor.rowcount, count)
 
     def test_rowcount_nodata(self):
         """
@@ -910,7 +1063,7 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(i int)")
         # This is a different code path internally.
         self.cursor.execute("delete from t1")
-        self.assertEquals(self.cursor.rowcount, 0)
+        self.assertEqual(self.cursor.rowcount, 0)
 
     def test_rowcount_select(self):
         """
@@ -925,23 +1078,25 @@ class SqlServerTestCase(unittest.TestCase):
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
         self.cursor.execute("select * from t1")
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, -1)
 
         rows = self.cursor.fetchall()
-        self.assertEquals(len(rows), count)
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(len(rows), count)
+        self.assertEqual(self.cursor.rowcount, -1)
 
     def test_rowcount_reset(self):
-        "Ensure rowcount is reset to -1"
+        "Ensure rowcount is reset after DDL"
+        
+        ddl_rowcount = 0 if self.driver_type_is('freetds') else -1
 
         self.cursor.execute("create table t1(i int)")
         count = 4
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
-        self.assertEquals(self.cursor.rowcount, 1)
+        self.assertEqual(self.cursor.rowcount, 1)
 
         self.cursor.execute("create table t2(i int)")
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, ddl_rowcount)
 
     #
     # always return Cursor
@@ -955,7 +1110,7 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(i int)")
         self.cursor.execute("insert into t1 values (1)")
         v = self.cursor.execute("delete from t1")
-        self.assertEquals(v, self.cursor)
+        self.assertEqual(v, self.cursor)
 
     def test_retcursor_nodata(self):
         """
@@ -967,13 +1122,13 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(i int)")
         # This is a different code path internally.
         v = self.cursor.execute("delete from t1")
-        self.assertEquals(v, self.cursor)
+        self.assertEqual(v, self.cursor)
 
     def test_retcursor_select(self):
         self.cursor.execute("create table t1(i int)")
         self.cursor.execute("insert into t1 values (1)")
         v = self.cursor.execute("select * from t1")
-        self.assertEquals(v, self.cursor)
+        self.assertEqual(v, self.cursor)
 
     #
     # misc
@@ -987,7 +1142,7 @@ class SqlServerTestCase(unittest.TestCase):
             self.cursor.execute("insert into [test one] values(1)")
             self.cursor.execute("select * from [test one]")
             v = self.cursor.fetchone()[0]
-            self.assertEquals(v, 1)
+            self.assertEqual(v, 1)
         finally:
             self.cnxn.rollback()
 
@@ -1005,7 +1160,7 @@ class SqlServerTestCase(unittest.TestCase):
         names = [ t[0] for t in self.cursor.description ]
         names.sort()
 
-        self.assertEquals(names, [ "abc", "def" ])
+        self.assertEqual(names, [ "abc", "def" ])
 
         # Put it back so other tests don't fail.
         pyodbc.lowercase = False
@@ -1021,7 +1176,7 @@ class SqlServerTestCase(unittest.TestCase):
 
         row = self.cursor.execute("select * from t1").fetchone()
 
-        self.assertEquals(self.cursor.description, row.cursor_description)
+        self.assertEqual(self.cursor.description, row.cursor_description)
 
 
     def test_temp_select(self):
@@ -1084,6 +1239,18 @@ class SqlServerTestCase(unittest.TestCase):
             self.assertEqual(param[0], row[0])
             self.assertEqual(param[1], row[1])
 
+    def test_executemany_dae_0(self):
+        """
+        DAE for 0-length value
+        """
+        self.cursor.execute("create table t1(a nvarchar(max))")
+
+        self.cursor.fast_executemany = True
+        self.cursor.executemany("insert into t1(a) values(?)", [['']])
+
+        self.assertEqual(self.cursor.execute("select a from t1").fetchone()[0], '')
+
+        self.cursor.fast_executemany = False
 
     def test_executemany_failure(self):
         """
@@ -1095,7 +1262,7 @@ class SqlServerTestCase(unittest.TestCase):
                    ('error', 'not an int'),
                    (3, 'good') ]
 
-        self.failUnlessRaises(pyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)
+        self.assertRaises(pyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)
 
 
     def test_row_slicing(self):
@@ -1105,13 +1272,13 @@ class SqlServerTestCase(unittest.TestCase):
         row = self.cursor.execute("select * from t1").fetchone()
 
         result = row[:]
-        self.failUnless(result is row)
+        self.assertTrue(result is row)
 
         result = row[:-1]
         self.assertEqual(result, (1,2,3))
 
         result = row[0:4]
-        self.failUnless(result is row)
+        self.assertTrue(result is row)
 
 
     def test_row_repr(self):
@@ -1153,8 +1320,8 @@ class SqlServerTestCase(unittest.TestCase):
         # Select from the view
         self.cursor.execute("select * from t2")
         rows = self.cursor.fetchall()
-        self.assert_(rows is not None)
-        self.assert_(len(rows) == 3)
+        self.assertTrue(rows is not None)
+        self.assertTrue(len(rows) == 3)
 
     def test_autocommit(self):
         self.assertEqual(self.cnxn.autocommit, False)
@@ -1333,7 +1500,23 @@ class SqlServerTestCase(unittest.TestCase):
         self.assertEqual(row.n, 1)
         self.assertEqual(type(row.blob), bytearray)
 
-        self.cursor.execute("update t1 set n=?, blob=?", 2, None)
+        sql = "update t1 set n=?, blob=?"
+        try:
+            self.cursor.execute(sql, 2, None)
+        except pyodbc.DataError:
+            if self.driver_type_is('freetds'):
+                # FREETDS_KNOWN_ISSUE
+                #
+                # cnxn.getinfo(pyodbc.SQL_DESCRIBE_PARAMETER) returns False for FreeTDS, so
+                # pyodbc can't call SQLDescribeParam to get the correct parameter type.
+                # This can lead to errors being returned from SQL Server when sp_prepexec is called, 
+                # e.g., "Implicit conversion from data type varchar to varbinary(max) is not allowed." 
+                #
+                # So at least verify that the user can manually specify the parameter type
+                self.cursor.setinputsizes([(), (pyodbc.SQL_VARBINARY, None, None)])
+                self.cursor.execute(sql, 2, None)
+            else:
+                raise
         row = self.cursor.execute("select * from t1").fetchone()
         self.assertEqual(row.n, 2)
         self.assertEqual(row.blob, None)
@@ -1391,11 +1574,11 @@ class SqlServerTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values (1, 'test1')")
         self.cursor.execute("insert into t1 values (1, 'test2')")
         rows = self.cursor.execute("select n, s from t1 order by s").fetchall()
-        self.assert_(rows[0] < rows[1])
-        self.assert_(rows[0] <= rows[1])
-        self.assert_(rows[1] > rows[0])
-        self.assert_(rows[1] >= rows[0])
-        self.assert_(rows[0] != rows[1])
+        self.assertTrue(rows[0] < rows[1])
+        self.assertTrue(rows[0] <= rows[1])
+        self.assertTrue(rows[1] > rows[0])
+        self.assertTrue(rows[1] >= rows[0])
+        self.assertTrue(rows[0] != rows[1])
 
         rows = list(rows)
         rows.sort() # uses <
@@ -1415,8 +1598,8 @@ class SqlServerTestCase(unittest.TestCase):
         cursor = None
 
         rows = self.cursor.execute("select n from t1").fetchall()
-        self.assertEquals(len(rows), 1)
-        self.assertEquals(rows[0][0], 1)
+        self.assertEqual(len(rows), 1)
+        self.assertEqual(rows[0][0], 1)
 
 
     def test_context_manager_fail(self):
@@ -1438,7 +1621,7 @@ class SqlServerTestCase(unittest.TestCase):
         cursor = None
 
         count = self.cursor.execute("select count(*) from t1").fetchone()[0]
-        self.assertEquals(count, 0)
+        self.assertEqual(count, 0)
 
 
     def test_cursor_context_manager_success(self):
@@ -1454,8 +1637,8 @@ class SqlServerTestCase(unittest.TestCase):
         cursor = None
 
         rows = self.cursor.execute("select n from t1").fetchall()
-        self.assertEquals(len(rows), 1)
-        self.assertEquals(rows[0][0], 1)
+        self.assertEqual(len(rows), 1)
+        self.assertEqual(rows[0][0], 1)
 
 
     def test_cursor_context_manager_fail(self):
@@ -1475,7 +1658,7 @@ class SqlServerTestCase(unittest.TestCase):
         cursor = None
 
         count = self.cursor.execute("select count(*) from t1").fetchone()[0]
-        self.assertEquals(count, 0)
+        self.assertEqual(count, 0)
 
 
     def test_untyped_none(self):
@@ -1501,7 +1684,7 @@ class SqlServerTestCase(unittest.TestCase):
                             ''')
         self.cnxn.commit()
         value = self.cursor.execute("select * from func1(?)", 'test').fetchone()[0]
-        self.assertEquals(value, 'test')
+        self.assertEqual(value, 'test')
 
     def test_no_fetch(self):
         # Issue 89 with FreeTDS: Multiple selects (or catalog functions that issue selects) without fetches seem to
@@ -1513,11 +1696,11 @@ class SqlServerTestCase(unittest.TestCase):
     def test_drivers(self):
         drivers = pyodbc.drivers()
         self.assertEqual(list, type(drivers))
-        self.assert_(len(drivers) > 1)
+        self.assertTrue(len(drivers) > 0)
 
-        m = re.search('DRIVER={([^}]+)}', self.connection_string, re.IGNORECASE)
+        m = re.search('DRIVER={?([^}]+?)}?;', self.connection_string, re.IGNORECASE)
         current = m.group(1)
-        self.assert_(current in drivers)
+        self.assertTrue(current in drivers)
 
     def test_prepare_cleanup(self):
         # When statement is prepared, it is kept in case the next execute uses the same statement.  This must be
@@ -1536,7 +1719,151 @@ class SqlServerTestCase(unittest.TestCase):
         # This is really making sure we are properly encoding and comparing the SQLSTATEs.
         self.cursor.execute("create table t1(s1 varchar(10) primary key)")
         self.cursor.execute("insert into t1 values ('one')")
-        self.failUnlessRaises(pyodbc.IntegrityError, self.cursor.execute, "insert into t1 values ('one')")
+        self.assertRaises(pyodbc.IntegrityError, self.cursor.execute, "insert into t1 values ('one')")
+
+    def test_emoticons_as_parameter(self):
+        # https://github.com/mkleehammer/pyodbc/issues/423
+        #
+        # When sending a varchar parameter, pyodbc is supposed to set ColumnSize to the number
+        # of characters.  Ensure it works even with 4-byte characters.
+        #
+        # http://www.fileformat.info/info/unicode/char/1f31c/index.htm
+
+        v = "x \U0001F31C z"
+
+        self.cursor.execute("create table t1(s varchar(100))")
+        self.cursor.execute("insert into t1 values (?)", v)
+
+        result = self.cursor.execute("select s from t1").fetchone()[0]
+
+        self.assertEqual(result, v)
+
+    def test_emoticons_as_literal(self):
+        # similar to `test_emoticons_as_parameter`, above, except for Unicode literal
+        #
+        # http://www.fileformat.info/info/unicode/char/1f31c/index.htm
+
+        v = "x \U0001F31C z"
+
+        self.cursor.execute("create table t1(s varchar(100))")
+        self.cursor.execute("insert into t1 values (N'%s')" % v)
+
+        result = self.cursor.execute("select s from t1").fetchone()[0]
+
+        self.assertEqual(result, v)
+        
+    def test_tvp(self):
+        # https://github.com/mkleehammer/pyodbc/issues/290
+        #
+        # pyodbc supports queries with table valued parameters in sql server
+        #
+
+        if self.handle_known_issues_for('freetds', print_reminder=True):
+            warn('FREETDS_KNOWN_ISSUE - test_tvp: test cancelled.')
+            return
+
+        # (Don't use "if exists" since older SQL Servers don't support it.)
+        try:
+            self.cursor.execute("drop procedure SelectTVP")
+        except:
+            pass
+        try:
+            self.cursor.execute("drop type TestTVP")
+        except:
+            pass
+        self.cursor.commit()
+        
+        query = "CREATE TYPE TestTVP AS TABLE("\
+                "c01 VARCHAR(255),"\
+                "c02 VARCHAR(MAX),"\
+                "c03 VARBINARY(255),"\
+                "c04 VARBINARY(MAX),"\
+                "c05 BIT,"\
+                "c06 DATE,"\
+                "c07 TIME,"\
+                "c08 DATETIME2(5),"\
+                "c09 BIGINT,"\
+                "c10 FLOAT,"\
+                "c11 NUMERIC(38, 24),"\
+                "c12 UNIQUEIDENTIFIER)"
+
+        self.cursor.execute(query)
+        self.cursor.commit()
+        self.cursor.execute("CREATE PROCEDURE SelectTVP @TVP TestTVP READONLY AS SELECT * FROM @TVP;")
+        self.cursor.commit()
+
+        long_string = ''
+        long_bytearray = []
+        for i in range(255):
+            long_string += chr((i % 95) + 32)
+            long_bytearray.append(i % 255)
+            
+        very_long_string = ''
+        very_long_bytearray = []
+        for i in range(2000000):
+            very_long_string += chr((i % 95) + 32)
+            very_long_bytearray.append(i % 255)
+            
+        c01 = ['abc', '', long_string]
+
+        c02 = ['abc', '', very_long_string]
+
+        c03 = [bytearray([0xD1, 0xCE, 0xFA, 0xCE]),
+               bytearray([0x00, 0x01, 0x02, 0x03, 0x04]),
+               bytearray(long_bytearray)]
+               
+        c04 = [bytearray([0x0F, 0xF1, 0xCE, 0xCA, 0xFE]),
+               bytearray([0x00, 0x01, 0x02, 0x03, 0x04, 0x05]),
+               bytearray(very_long_bytearray)]
+
+        c05 = [1, 0, 1]
+
+        c06 = [date(1997, 8, 29),
+               date(1, 1, 1),
+               date(9999, 12, 31)]
+               
+        c07 = [time(9, 13, 39),
+               time(0, 0, 0),
+               time(23, 59, 59)]
+               
+        c08 = [datetime(2018, 11, 13, 13, 33, 26, 298420),
+               datetime(1, 1, 1, 0, 0, 0, 0),
+               datetime(9999, 12, 31, 23, 59, 59, 999990)]
+               
+        c09 = [1234567, -9223372036854775808, 9223372036854775807]
+
+        c10 = [3.14, -1.79E+308, 1.79E+308]
+
+        c11 = [Decimal('31234567890123.141243449787580175325274'),
+               Decimal(             '0.000000000000000000000001'),
+               Decimal('99999999999999.999999999999999999999999')]
+
+        c12 = ['4FE34A93-E574-04CC-200A-353F0D1770B1',
+               '33F7504C-2BAC-1B83-01D1-7434A7BA6A17',
+               'FFFFFFFF-FFFF-FFFF-FFFF-FFFFFFFFFFFF']
+
+        param_array = []       
+
+        for i in range (3):
+            param_array.append([c01[i], c02[i], c03[i], c04[i], c05[i], c06[i], c07[i], c08[i], c09[i], c10[i], c11[i], c12[i]])
+
+        success = True
+
+        try:
+            result_array = self.cursor.execute("exec SelectTVP ?",[param_array]).fetchall()
+        except Exception as ex:
+            print("Failed to execute SelectTVP")
+            print("Exception: [" + type(ex).__name__ + "]" , ex.args)
+            
+            success = False
+        else:
+            for r in range(len(result_array)):
+                for c in range(len(result_array[r])):
+                    if(result_array[r][c] != param_array[r][c]):
+                        print("Mismatch at row " + str(r+1) + ", column " + str(c+1) + "; expected:", param_array[r][c] , " received:", result_array[r][c])
+                        success = False
+
+        self.assertEqual(success, True)
 
 
 def main():
@@ -1569,6 +1896,8 @@ def main():
     testRunner = unittest.TextTestRunner(verbosity=options.verbose)
     result = testRunner.run(suite)
 
+    return result
+
 
 if __name__ == '__main__':
 
@@ -1577,4 +1906,4 @@ if __name__ == '__main__':
     add_to_path()
 
     import pyodbc
-    main()
+    sys.exit(0 if main().wasSuccessful() else 1)
