@@ -1,9 +1,25 @@
 #!/usr/bin/env python
+# -*- coding: utf-8 -*-
 
-# Unit tests for PostgreSQL on OS X and Linux.
+usage = """\
+usage: %prog [options] connection_string
 
-from __future__ import print_function
+Unit tests for PostgreSQL.  To use, pass a connection string as the parameter.
+The tests will create and drop tables t1 and t2 as necessary.
 
+These run using the version from the 'build' directory, not the version
+installed into the Python directories.  You must run python setup.py build
+before running the tests.
+
+You can also put the connection string into a tmp/setup.cfg file like so:
+
+  [pgtests]
+  connection-string=DSN=PostgreSQL35W
+
+Note: Be sure to use the "Unicode" (not the "ANSI") version of the PostgreSQL ODBC driver.
+"""
+
+import sys
 import uuid
 import unittest
 from decimal import Decimal
@@ -90,11 +106,11 @@ class PGTestCase(unittest.TestCase):
 
     def test_drivers(self):
         p = pyodbc.drivers()
-        self.assert_(isinstance(p, list))
+        self.assertTrue(isinstance(p, list))
 
     def test_datasources(self):
         p = pyodbc.dataSources()
-        self.assert_(isinstance(p, dict))
+        self.assertTrue(isinstance(p, dict))
 
     # def test_gettypeinfo(self):
     #     self.cursor.getTypeInfo(pyodbc.SQL_VARCHAR)
@@ -106,19 +122,19 @@ class PGTestCase(unittest.TestCase):
 
     def test_getinfo_string(self):
         value = self.cnxn.getinfo(pyodbc.SQL_CATALOG_NAME_SEPARATOR)
-        self.assert_(isinstance(value, str))
+        self.assertTrue(isinstance(value, str))
 
     def test_getinfo_bool(self):
         value = self.cnxn.getinfo(pyodbc.SQL_ACCESSIBLE_TABLES)
-        self.assert_(isinstance(value, bool))
+        self.assertTrue(isinstance(value, bool))
 
     def test_getinfo_int(self):
         value = self.cnxn.getinfo(pyodbc.SQL_DEFAULT_TXN_ISOLATION)
-        self.assert_(isinstance(value, int))
+        self.assertTrue(isinstance(value, int))
 
     def test_getinfo_smallint(self):
         value = self.cnxn.getinfo(pyodbc.SQL_CONCAT_NULL_BEHAVIOR)
-        self.assert_(isinstance(value, int))
+        self.assertTrue(isinstance(value, int))
 
 
     def test_negative_float(self):
@@ -330,20 +346,20 @@ class PGTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(s varchar(20))")
         self.cursor.execute("insert into t1 values(?)", "1")
         row = self.cursor.execute("select * from t1").fetchone()
-        self.assertEquals(row[0], "1")
-        self.assertEquals(row[-1], "1")
+        self.assertEqual(row[0], "1")
+        self.assertEqual(row[-1], "1")
 
     def test_version(self):
-        self.assertEquals(3, len(pyodbc.version.split('.'))) # 1.3.1 etc.
+        self.assertEqual(3, len(pyodbc.version.split('.'))) # 1.3.1 etc.
 
     def test_rowcount_delete(self):
-        self.assertEquals(self.cursor.rowcount, -1)
+        self.assertEqual(self.cursor.rowcount, -1)
         self.cursor.execute("create table t1(i int)")
         count = 4
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
         self.cursor.execute("delete from t1")
-        self.assertEquals(self.cursor.rowcount, count)
+        self.assertEqual(self.cursor.rowcount, count)
 
     def test_rowcount_nodata(self):
         """
@@ -356,7 +372,7 @@ class PGTestCase(unittest.TestCase):
         self.cursor.execute("create table t1(i int)")
         # This is a different code path internally.
         self.cursor.execute("delete from t1")
-        self.assertEquals(self.cursor.rowcount, 0)
+        self.assertEqual(self.cursor.rowcount, 0)
 
     def test_rowcount_select(self):
         self.cursor.execute("create table t1(i int)")
@@ -364,7 +380,7 @@ class PGTestCase(unittest.TestCase):
         for i in range(count):
             self.cursor.execute("insert into t1 values (?)", i)
         self.cursor.execute("select * from t1")
-        self.assertEquals(self.cursor.rowcount, 4)
+        self.assertEqual(self.cursor.rowcount, 4)
 
     # PostgreSQL driver fails here?
     # def test_rowcount_reset(self):
@@ -374,10 +390,10 @@ class PGTestCase(unittest.TestCase):
     #     count = 4
     #     for i in range(count):
     #         self.cursor.execute("insert into t1 values (?)", i)
-    #     self.assertEquals(self.cursor.rowcount, 1)
+    #     self.assertEqual(self.cursor.rowcount, 1)
     #
     #     self.cursor.execute("create table t2(i int)")
-    #     self.assertEquals(self.cursor.rowcount, -1)
+    #     self.assertEqual(self.cursor.rowcount, -1)
 
     def test_lower_case(self):
         "Ensure pyodbc.lowercase forces returned column names to lowercase."
@@ -393,7 +409,7 @@ class PGTestCase(unittest.TestCase):
         names = [ t[0] for t in self.cursor.description ]
         names.sort()
 
-        self.assertEquals(names, [ "abc", "def" ])
+        self.assertEqual(names, [ "abc", "def" ])
 
         # Put it back so other tests don't fail.
         pyodbc.lowercase = False
@@ -408,7 +424,7 @@ class PGTestCase(unittest.TestCase):
         self.cursor.execute("insert into t1 values(1, 'abc')")
 
         row = self.cursor.execute("select * from t1").fetchone()
-        self.assertEquals(self.cursor.description, row.cursor_description)
+        self.assertEqual(self.cursor.description, row.cursor_description)
 
 
     def test_executemany(self):
@@ -466,7 +482,7 @@ class PGTestCase(unittest.TestCase):
                    ('error', 'not an int'),
                    (3, 'good') ]
 
-        self.failUnlessRaises(pyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)
+        self.assertRaises(pyodbc.Error, self.cursor.executemany, "insert into t1(a, b) value (?, ?)", params)
 
 
     def test_row_slicing(self):
@@ -476,13 +492,13 @@ class PGTestCase(unittest.TestCase):
         row = self.cursor.execute("select * from t1").fetchone()
 
         result = row[:]
-        self.failUnless(result is row)
+        self.assertTrue(result is row)
 
         result = row[:-1]
         self.assertEqual(result, (1,2,3))
 
         result = row[0:4]
-        self.failUnless(result is row)
+        self.assertTrue(result is row)
 
     def test_cnxn_execute_error(self):
         """
@@ -492,7 +508,7 @@ class PGTestCase(unittest.TestCase):
         """
         self.cursor.execute("create table t1(a int primary key)")
         self.cursor.execute("insert into t1 values (1)")
-        self.failUnlessRaises(pyodbc.Error, self.cnxn.execute, "insert into t1 values (1)")
+        self.assertRaises(pyodbc.Error, self.cnxn.execute, "insert into t1 values (1)")
 
     def test_row_repr(self):
         self.cursor.execute("create table t1(a int, b int, c int, d int)")
@@ -522,7 +538,7 @@ class PGTestCase(unittest.TestCase):
         # This is really making sure we are properly encoding and comparing the SQLSTATEs.
         self.cursor.execute("create table t1(s1 varchar(10) primary key)")
         self.cursor.execute("insert into t1 values ('one')")
-        self.failUnlessRaises(pyodbc.IntegrityError, self.cursor.execute, "insert into t1 values ('one')")
+        self.assertRaises(pyodbc.IntegrityError, self.cursor.execute, "insert into t1 values ('one')")
 
 
     def test_cnxn_set_attr_before(self):
@@ -548,7 +564,7 @@ class PGTestCase(unittest.TestCase):
         # I'm not sure why, but PyArg_ParseTupleAndKeywords fails if you use "|s" for an
         # optional string keyword when calling indirectly.
 
-        self.cursor.execute("create table t1(a int, b varchar(3))")
+        self.cursor.execute("create table t1(a int, b varchar(3), xΏz varchar(4))")
 
         self.cursor.columns('t1')
         results = {row.column_name: row for row in self.cursor}
@@ -557,6 +573,9 @@ class PGTestCase(unittest.TestCase):
         row = results['b']
         assert row.type_name == 'varchar'
         assert row.precision == 3, row.precision
+        row = results['xΏz']
+        assert row.type_name == 'varchar'
+        assert row.precision == 4, row.precision
 
         # Now do the same, but specifically pass in None to one of the keywords.  Old versions
         # were parsing arguments incorrectly and would raise an error.  (This crops up when
@@ -576,7 +595,73 @@ class PGTestCase(unittest.TestCase):
         self.cursor.execute("select 1")
         self.cursor.cancel()
 
+    def test_emoticons_as_parameter(self):
+        # https://github.com/mkleehammer/pyodbc/issues/423
+        #
+        # When sending a varchar parameter, pyodbc is supposed to set ColumnSize to the number
+        # of characters.  Ensure it works even with 4-byte characters.
+        #
+        # http://www.fileformat.info/info/unicode/char/1f31c/index.htm
 
+        v = "x \U0001F31C z"
+
+        self.cursor.execute("CREATE TABLE t1(s varchar(100))")
+        self.cursor.execute("insert into t1 values (?)", v)
+
+        result = self.cursor.execute("select s from t1").fetchone()[0]
+
+        self.assertEqual(result, v)
+
+    def test_emoticons_as_literal(self):
+        # https://github.com/mkleehammer/pyodbc/issues/630
+
+        v = "x \U0001F31C z"
+
+        self.cursor.execute("CREATE TABLE t1(s varchar(100))")
+        self.cursor.execute("insert into t1 values ('%s')" % v)
+
+        result = self.cursor.execute("select s from t1").fetchone()[0]
+
+        self.assertEqual(result, v)
+        
+    def test_output_conversion(self):
+        # Note the use of SQL_WVARCHAR, not SQL_VARCHAR.
+
+        def convert(value):
+            # The value is the raw bytes (as a bytes object) read from the
+            # database.  We'll simply add an X at the beginning at the end.
+            return 'X' + value.decode('latin1') + 'X'
+
+        self.cursor.execute("create table t1(n int, v varchar(10))")
+        self.cursor.execute("insert into t1 values (1, '123.45')")
+
+        self.cnxn.add_output_converter(pyodbc.SQL_WVARCHAR, convert)
+        value = self.cursor.execute("select v from t1").fetchone()[0]
+        self.assertEqual(value, 'X123.45X')
+
+        # Clear all conversions and try again.  There should be no Xs this time.
+        self.cnxn.clear_output_converters()
+        value = self.cursor.execute("select v from t1").fetchone()[0]
+        self.assertEqual(value, '123.45')
+
+        # Same but clear using remove_output_converter.
+        self.cnxn.add_output_converter(pyodbc.SQL_WVARCHAR, convert)
+        value = self.cursor.execute("select v from t1").fetchone()[0]
+        self.assertEqual(value, 'X123.45X')
+
+        self.cnxn.remove_output_converter(pyodbc.SQL_WVARCHAR)
+        value = self.cursor.execute("select v from t1").fetchone()[0]
+        self.assertEqual(value, '123.45')
+
+        # And lastly, clear by passing None for the converter.
+        self.cnxn.add_output_converter(pyodbc.SQL_WVARCHAR, convert)
+        value = self.cursor.execute("select v from t1").fetchone()[0]
+        self.assertEqual(value, 'X123.45X')
+
+        self.cnxn.add_output_converter(pyodbc.SQL_WVARCHAR, None)
+        value = self.cursor.execute("select v from t1").fetchone()[0]
+        self.assertEqual(value, '123.45')
+        
 def main():
     from optparse import OptionParser
     parser = OptionParser(usage="usage: %prog [options] connection_string")
@@ -618,7 +703,10 @@ def main():
         s = unittest.TestSuite([ PGTestCase(connection_string, options.ansi, m) for m in methods ])
 
     testRunner = unittest.TextTestRunner(verbosity=options.verbose)
-    testRunner.run(s)
+    result = testRunner.run(s)
+
+    return result
+
 
 if __name__ == '__main__':
 
@@ -627,4 +715,4 @@ if __name__ == '__main__':
     add_to_path()
 
     import pyodbc
-    main()
+    sys.exit(0 if main().wasSuccessful() else 1)
